#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External
from os.path import dirname, join

# INternal
from dbgen import Const, Env, Generator, Import, Model, PyBlock, __file__
from dbgen.example.constants import defaultEnv
from dbgen.example.scripts.parsers import get_electrode, parse_expt, parse_proc_csv, parse_sqlite, parse_ssn

############################################################################
############################################################################
########################################################################
root = join(dirname(__file__), "example/data/")

################################################################################
############################################################################
############################################################################


def io(model: Model) -> None:

    tabs = [
        "sample",
        "scientist",
        "procedures",
        "history_detail",
        "history",
        "electrode",
        "anode",
        "cathode",
        "fuel_cell",
    ]

    (
        Sample,
        Scientist,
        Procedures,
        History_details,
        History,
        Electrode,
        Anode,
        Cathode,
        Fuel_cell,
    ) = map(model.get, tabs)

    ############################################################################
    ############################################################################

    pb1 = PyBlock(
        func=parse_ssn,
        args=[Const("ssn.json")],
        env=defaultEnv,
        outnames=["firstname", "lastname", "ssn"],
    )

    scientists = Generator(
        name="scientists",
        desc="populates Scientist table",
        transforms=[pb1],
        loads=[
            Scientist(
                insert=True,
                ssn=pb1["ssn"],
                firstname=pb1["firstname"],
                lastname=pb1["lastname"],
            )
        ],
    )

    ############################################################################

    dd_env = Env([Import("collections", "defaultdict"), Import("csv", "DictReader")])

    ghcpb = PyBlock(
        func=parse_proc_csv,
        env=defaultEnv + dd_env,
        args=[Const("procedures.csv")],
        outnames=[
            "id",
            "step",
            "procedure_name",
            "timestamp",
            "ssn",
            "value",
            "dtype",
            "name",
        ],
    )

    sample_load = Sample(insert=True, id=ghcpb["id"])

    proc_load = Procedures(insert=True, procedure_name=ghcpb["procedure_name"])

    sci_load = Scientist(insert=True, ssn=ghcpb["ssn"])

    hist_load = History(
        insert=True,
        step=ghcpb["step"],
        sample=sample_load,
        expt_type=proc_load,
        operator=sci_load,
    )

    histd_load = History_details(
        insert=True,
        name=ghcpb["name"],
        value=ghcpb["value"],
        dtype=ghcpb["dtype"],
        history=hist_load,
    )
    get_history_csv = Generator(
        name="get_history_csv",
        desc="Parse CSV file with History data",
        transforms=[ghcpb],
        loads=[histd_load],
    )

    ############################################################################

    ca = []

    for x in ["Cathode", "Anode"]:
        capb = PyBlock(
            func=get_electrode,
            env=defaultEnv,
            args=[Const("experiment.json"), Const(x)],
            outnames=["id", "expt_id", "comp"],
        )

        samp_load = Sample(insert=True, id=capb["id"])

        elec_load = Electrode(insert=True, composition=capb["comp"], sample=samp_load)

        x_load = model[x](insert=True, electrode=elec_load)

        fc_load = Fuel_cell(insert=True, **{x: x_load, "expt_id": capb["expt_id"]})
        ca.append(
            Generator(
                name=x,
                desc=f"Extract {x} info from an experiment.json",
                transforms=[capb],
                loads=[fc_load],
            )
        )

    cathode, anode = ca
    ############################################################################
    sqlite_env = Env([Import("sqlite3", "connect")])

    ghd = PyBlock(
        parse_sqlite,
        env=defaultEnv + sqlite_env,
        args=[Const("procedure.db")],
        outnames=["id", "step", "pname", "fname", "lname", "ssn"],
    )

    samact = Sample(insert=True, id=ghd["id"])

    sciact = Scientist(
        insert=True,
        ssn=ghd["ssn"],
        firstname=ghd["fname"],
        lastname=ghd["lname"],
    )

    proact = Procedures(insert=True, procedure_name=ghd["pname"])

    hact = History(
        insert=True,
        step=ghd["step"],
        sample=samact,
        expt_type=proact,
        operator=sciact,
    )

    get_history_db = Generator(
        name="get_history_db",
        desc="Parse SQLite file with History data",
        transforms=[ghd],
        loads=[hact],
    )
    ############################################################################
    details = ["expt_id", "timestamp", "capacity", "electrolyte"]
    fd_pb = PyBlock(parse_expt, env=defaultEnv, args=[Const("experiment.json")], outnames=details)
    fuel_details = Generator(
        name="fuel_details",
        desc="other details about fuel cell experiments",
        loads=[Fuel_cell(insert=True, **{x: fd_pb[x] for x in details})],
        transforms=[fd_pb],
    )

    ############################################################################
    ############################################################################

    model.add(
        [
            scientists,
            get_history_csv,
            cathode,
            anode,
            get_history_db,
            fuel_details,
        ]
    )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External
from copy import deepcopy
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import List as L
from typing import Set as S
from typing import Union as U

from tqdm import tqdm

from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.core.schema import Attr, AttrTup, Entity, Path, PathEQ, QView, RawView, Rel, RelTup, View

# Internal
from dbgen.utils.misc import Base
from dbgen.utils.sql import sqlexecute, sqlselect

if TYPE_CHECKING:
    from networkx import DiGraph
############################################################################


class Schema(Base):
    """Unnamed collection of Objects, Views, Relations, Path Equations"""

    def __init__(
        self,
        objlist: L[Entity] = None,
        viewlist: L[View] = None,
        pes: L[PathEQ] = None,
    ) -> None:

        self.objlist = objlist or []
        self.viewlist = viewlist or []
        from networkx import DiGraph

        self._fks = DiGraph()
        self._fks.add_nodes_from(self.objlist)  # nodes are object NAMES

        self.pes = set(pes or [])  # path equivalencies

        for o in self.objlist:
            for rel in o.fks:
                self._add_relation(rel.to_rel(o.name))

        super().__init__()

    def __str__(self) -> str:
        return "Schema<%d objs, %d rels>" % (len(self.objlist), len(self._fks))

    def __getitem__(self, key: str) -> Entity:
        # raise NotImplementedError
        entity = self.objs[key.lower()]
        # if partition_value:
        #     return entity.get_partition(partition_value)
        return entity

    def get_entity(self, key: str, partition_value: Any) -> Entity:
        entity = self.objs[key.lower()]
        if partition_value:
            return entity.get_partition(partition_value)
        return entity

    def __contains__(self, key: str) -> bool:
        return key.lower() in self.objs

    @property
    def objs(self) -> D[str, Entity]:
        output: D[str, Entity] = {}
        for o in self.objlist:
            output[o.name] = o
            if o.is_partitioned:
                output.update({part.name: part for part in o.get_all_partitions()})
        return output

    @property
    def views(self) -> D[str, View]:
        return {o.name.lower(): o for o in self.viewlist}

    def make_schema(self, conn: ConnI, nuke: bool = False, bar: bool = True) -> None:
        """Create empty schema."""
        if nuke:
            safe_conn = deepcopy(conn)
            safe_cxn = safe_conn.connect()
            sqlexecute(safe_cxn, f'DROP SCHEMA IF EXISTS "{conn.schema}" CASCADE')
            sqlexecute(safe_cxn, f'CREATE SCHEMA "{conn.schema}"')
        cxn = conn.connect()
        tqargs = dict(leave=False, disable=not bar)

        for ta in tqdm(self.objs.values(), desc="adding tables", **tqargs):
            if not getattr(ta, "_is_view", False):
                for sql in ta.create():
                    sqlexecute(cxn, sql)
        for r in tqdm(self._rels(), desc="adding relations", **tqargs):
            sqlexecute(cxn, self._create_fk(r))

        for vn in tqdm(self.views, desc="adding views", **tqargs):
            q = f'CREATE OR REPLACE VIEW "{vn}" AS {self._show_view(vn)};'
            sqlexecute(cxn, q)

    def check_schema_exists(self, conn: ConnI) -> bool:
        cxn = conn.connect()
        q = f"SELECT table_name from information_schema.tables where \
        table_schema = '{conn.schema}'"
        tables_in_db = [x[0] for x in sqlselect(cxn, q)]
        return all([obj in tables_in_db for obj in self.objs])

    ################
    # Adding stuff #
    ################

    def _add_attr(self, oname: str, a: Attr) -> None:
        """Add to model"""
        self[oname].attrs.append(a)

    def _add_object(self, o: Entity) -> None:
        """Add to model"""

        if o.name in self:  # Validate
            raise ValueError(f"Cannot add {o}, name is already taken!")
        else:
            self.objlist.append(o)  # Add
            self._fks.add_node(o.name)

        for rel in o.fkdict.values():
            self._add_relation(rel)

    def _add_view(self, v: View) -> None:
        """Add to model"""
        # Validate
        # --------
        if v.name in self.views or v.name in self.objs:
            raise ValueError(f"Cannot add {v}, name already taken!")
        # Add
        # ----
        self.viewlist.append(v)

    def _add_relation(self, r: Rel) -> None:
        """Add to model"""
        # Validate
        # ---------
        assert isinstance(r, Rel)
        err = "Cannot add %s: %s not found in model %s"
        assert r.o1 in self, err % (r, r.o1, self.objlist)
        assert r.o2 in self, err % (r, r.o2, self.objlist)

        notfound = "Cannot add %s, %s already has a %s with that name "

        for _, _, data in self._fks.edges(r.o1, data=True):
            if r.name in [fk.name for fk in data["fks"]]:
                raise ValueError(notfound % (r, r.o1, "relation"))
            if r.name in self[r.o1].attrnames():
                raise ValueError(notfound % (r, r.o1, "attribute"))

        # Add
        # ----
        self.add_fk(self._fks, r)

    def _add_patheq(self, peq: PathEQ) -> None:
        """Add to model"""
        # Validate
        # ---------
        rand = next(iter(peq))
        start = rand.start()
        end = rand._path_end(self)
        for p in peq:
            self._validate_path(p)
            assert p.start() == start
            assert p._path_end(self) == end

        self.pes.add(peq)

    @staticmethod
    def add_fk(G: "DiGraph", r: Rel, forward: bool = True) -> None:
        """ Modify a graph (used by both add_relation and info_graph) """
        a, b = (r.o1, r.o2)
        if not forward:
            a, b = b, a
        if G.has_edge(a, b):
            G[a][b]["fks"].append(r)
            G[a][b]["fks"] = sorted(G[a][b]["fks"], key=lambda x: repr(x))
        else:
            G.add_edge(a, b, fks=list([r]))

    ###########
    # Objects #
    ###########
    def add_cols(self, obj: Entity) -> L[str]:
        attr_stmts = []
        for c in obj.attrs:
            col_name, col_desc, c_index = c.create_col(obj.name)
            stmt = f"ALTER TABLE {obj.name} ADD COLUMN IF NOT EXISTS {col_name}"
            attr_stmts.append(stmt)
            attr_stmts.append(col_desc)
            if c_index:
                attr_stmts.append(c_index)
        rel_stmts = [self._create_fk(rel) for rel in obj.fkdict.values()]
        return attr_stmts + rel_stmts

    #########
    # Exprs #
    #########
    def all_attr(self) -> S[AttrTup]:
        out = set()
        for o in self.objs.values():
            for a in o.attrnames():
                out.add(o[a])
        return out

    ################
    # FK Graph #
    ################
    def _create_fk(self, fk: Rel) -> str:
        """create SQL FK statement"""
        # Check rel for a non-default trigger
        if fk.delete_trigger != "NO ACTION":
            trigger = f" ON DELETE {fk.delete_trigger}"
        else:
            trigger = ""
        args = [fk.o1, fk.name, fk.o2, self[fk.o2].id_str, trigger]
        stmt = 'ALTER TABLE "{0}" ADD COLUMN IF NOT EXISTS "{1}" BIGINT;'
        stmt += 'ALTER TABLE "{0}" DROP CONSTRAINT IF EXISTS fk__{0}__{1}__{2}__{3};'
        stmt += 'ALTER TABLE "{0}" ADD CONSTRAINT fk__{0}__{1}__{2}__{3}\
        FOREIGN KEY ("{1}") REFERENCES "{2}"("{3}") {4};'
        stmt += 'CREATE INDEX IF NOT EXISTS "{0}__{1}__fkey" ON "{0}"("{1}");'
        return stmt.format(*args)

    def _rels(self) -> S[Rel]:
        """ ALL relations between any objects, in some order """
        fks = set()
        for _, _, d in self._fks.edges(data=True):
            for fk in d["fks"]:
                fks.add(fk)
        return fks

    def _obj_all_fks(self, o: Entity) -> S[Rel]:
        """ Relations that start OR end on a given object """
        inward = set.union(*[d["fks"] for _, _, d in self._fks.in_edges(o.name, data=True)])
        return set(o.fkdict.values()) | inward

    def get_rel(self, r: U[Rel, RelTup]) -> Rel:
        """
        Upgrade a Relation representation that only has limited
        (but identifying) info
        """
        if isinstance(r, Rel):
            return r
        else:
            return self[r.obj].fkdict[r.rel]

    def _info_graph(self, links: L[U[Rel, RelTup]]) -> "DiGraph":
        """Natural paths of information propagation, which includes the normal
        Rel relationships but also taking into account a
        user-specified list of relationships that are allowed to propagate
        information in the 'reverse' direction

        Furthermore, 1-1 relationships are identified and information is
        allowed to propagate in opposite direction there, too.
        """
        G = self._fks.copy()

        for name, o in self.objs.items():
            pars = [fk for fk in o.fkdict.values() if fk.identifying]
            # only if it's a 1-1 table
            if len(pars) == 1 and len(o.ids()) == 0:
                p = pars[0]  # identifying foreign key
                self.add_fk(G, p, forward=False)

        for fk in links:
            if isinstance(fk, RelTup):
                fk = self.get_rel(fk)
            self.add_fk(G, fk, forward=False)

        return G

    #########
    # Paths #
    #########
    def _validate_path(self, p: Path) -> None:
        """Throw error if invalid path is passed"""
        curr = self[p.start()]
        for r in p.rels:
            rel = self.get_rel(r)
            assert rel.o1 == curr.name
            curr = self[rel.o2]
        if hasattr(p, "attr"):
            assert getattr(p, "attr").obj == curr.name

    #########
    # Views #
    #########
    def _show_view(self, vname: str) -> str:
        v = self.views[vname]
        if isinstance(v, RawView):
            return v.raw
        elif isinstance(v, QView):
            return v.q.showQ()
        else:
            raise TypeError



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Uncategorized utilities for core dbgen functionality"""
import re
from contextlib import suppress
from json import dump, load
from os import environ
from os.path import exists
from pprint import pformat

# External Modules
from typing import TYPE_CHECKING, Any
from typing import Callable as C
from typing import List as L
from typing import Tuple as T

from hypothesis.strategies import SearchStrategy, builds
from psycopg2 import Error, OperationalError, connect
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT, parse_dsn

from dbgen.utils.misc import Base

# Internal Modules
if TYPE_CHECKING:
    from airflow.hooks.connections import Connection
    from sshtunnel import SSHTunnelForwarder

    from dbgen.core.gen import Generator


class ConnectInfo(Base):
    """
    PostGreSQL connection info
    """

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 5432,
        user: str = None,
        passwd: str = None,
        db: str = "",
        schema: str = "public",
        ssh: str = "",
        ssh_port: int = 22,
        ssh_username: str = "",
        ssh_pkey: str = "",
        remote_bind_address: str = "localhost",
        remote_bind_port: int = 5432,
    ) -> None:
        """
        Create a ConnectInfo object to wrap around postgres connection information.

        Args:
            host (str, optional): hostname/ip address for postgres server. Defaults to "127.0.0.1".
            port (int, optional): port number postgres server is listening on. Defaults to 5432.
            user (str, optional): username to login to postgres with. Defaults to None.
            passwd (str, optional): password for above user. Defaults to None.
            db (str, optional): database to login to. Defaults to "".
            schema (str, optional): set the schema to search. Defaults to "public".
            ssh (str, optional): ssh hostname to tunnel through. Defaults to "".
            ssh_port (int, optional): port the host is using to listen to ssh connections. Defaults to 22.
            ssh_username (str, optional): username to log onto ssh tunnel. Defaults to "".
            ssh_pkey (str, optional): password to log onto the ssh tunnel. Defaults to "".
            remote_bind_address (str, optional): what host to query on ssh machine. Defaults to "localhost".
            remote_bind_port (int, optional): what port postgres is listening to
            on ssh machine. Defaults to 5432.
        """

        if not user:
            user = passwd = environ.get("USER", "")

        self.host = host
        self.port = port
        self.user = user
        self.passwd = passwd
        self.db = db
        self.schema = schema
        self.ssh = ssh
        self.ssh_port = ssh_port
        self.ssh_username = ssh_username
        self.ssh_pkey = ssh_pkey
        self.remote_bind_address = remote_bind_address
        self.remote_bind_port = remote_bind_port
        super().__init__()

    def __str__(self) -> str:
        return pformat(self.__dict__)

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    def tunnel(self) -> "SSHTunnelForwarder":
        from sshtunnel import SSHTunnelForwarder

        return (
            SSHTunnelForwarder(
                (self.ssh, self.ssh_port),
                ssh_username=self.ssh_username,
                ssh_pkey=self.ssh_pkey,
                remote_bind_address=(
                    self.remote_bind_address,
                    self.remote_bind_port,
                ),
            )
            if self.ssh
            else suppress()
        )

    def connect(self, attempt: int = 3, auto_commit: bool = True) -> "Connection":
        for _ in range(attempt):
            try:
                with self.tunnel():
                    conn = connect(
                        host=self.host,
                        port=self.port,
                        user=self.user,
                        password=self.passwd,
                        dbname=self.db,
                        connect_timeout=28800,
                    )
                    if auto_commit:
                        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
                    if self.schema != "public":
                        cur = conn.cursor()
                        cur.execute(f'SET search_path TO "{self.schema}"')
                    return conn
            except OperationalError as exc:
                if re.findall("database.*does not exist", str(exc)):
                    raise OperationalError(
                        f"Database {self.db} does not exist,"
                        "please check connection or create DB before running DBgen or first time"
                    )
                raise exc
            # except Error as exc:
            #     sleep(1)
        raise Error(
            f"Exceeded number of attempts to connect to host using credentials."
            "Please make sure the database is running and you have provided the correct credentials."
        )

    def to_file(self, pth: str) -> None:
        """Store connectinfo data as a JSON file"""
        with open(pth, "w") as f:
            dump(vars(self), f)

    @staticmethod
    def from_file(pth: str) -> "ConnectInfo":
        """
        Create from path to file with ConnectInfo fields in JSON format
        """
        assert exists(pth), "Error loading connection info: no file at " + pth
        with open(pth) as f:
            return ConnectInfo(**load(f))

    @staticmethod
    def from_dsn(dsn: str, schema: str = None) -> "ConnectInfo":
        """
        Create from path to file with ConnectInfo fields in JSON format
        """
        parsed_dsn = parse_dsn(dsn)
        kwargs = dict(
            user=parsed_dsn["user"],
            db=parsed_dsn["dbname"],
            host=parsed_dsn["host"],
            port=int(parsed_dsn.get("port", "5432")),
            passwd=parsed_dsn.get("password"),
        )
        if schema:
            kwargs["schema"] = schema
        return ConnectInfo(**kwargs)

    @staticmethod
    def from_postgres_hook(airflow_connection: "Connection") -> "ConnectInfo":
        """
        Create from path to file with ConnectInfo fields in JSON format
        """
        kwargs = dict(
            host=airflow_connection.host,
            port=airflow_connection.port,
            user=airflow_connection.login,
            passwd=airflow_connection.get_password(),
            db=airflow_connection.schema,
        )
        return ConnectInfo(**kwargs)

    def neutral(self) -> "Connection":
        copy = self.copy()
        copy.db = "postgres"
        conn = copy.connect()
        return conn.cursor()

    def kill(self) -> None:
        """Kills connections to the DB"""
        killQ = """SELECT pg_terminate_backend(pg_stat_activity.pid)
                    FROM pg_stat_activity
                    WHERE pg_stat_activity.datname = %s
                      AND pid <> pg_backend_pid();"""
        with self.neutral() as cxn:
            cxn.execute(killQ, vars=[self.db])

    def drop(self) -> None:
        """Completely removes a DB"""
        drop_stmt = f'DROP SCHEMA IF EXISTS "{self.schema}" CASCADE'
        conn = self.connect()
        with conn.cursor() as cxn:
            cxn.execute(drop_stmt)

    def create(self) -> None:
        """Kills connections to the DB"""
        create_stmt = f'CREATE SCHEMA IF NOT EXISTS "{self.schema}";'
        conn = self.connect()
        with conn.cursor() as cxn:
            cxn.execute(create_stmt)


################################################################################


class Dep(Base):
    """
    Capture dependency information between two Generators that modify a DB
    through four different sets: the tabs/cols that are inputs/outputs.
    """

    def __init__(
        self,
        tabs_needed: L[str] = [],
        cols_needed: L[str] = [],
        tabs_yielded: L[str] = [],
        cols_yielded: L[str] = [],
    ) -> None:
        """
        Initialize the Dep object with the tables and columns needed and yielded
        by a generator.

        Args:
            tabs_needed (L[str], optional): The tables the generator queries from. Defaults to [].
            cols_needed (L[str], optional): The columns the generator queries. Defaults to [].
            tabs_yielded (L[str], optional): The tables the generator loads into. Defaults to [].
            cols_yielded (L[str], optional): The columns the generator loads into. Defaults to [].
        """
        allts = [tabs_needed, tabs_yielded]
        allcs = [cols_needed, cols_yielded]
        assert all([all(["." not in t for t in ts]) for ts in allts]), allts
        assert all([all(["." in c for c in cs]) for cs in allcs]), allcs
        self.tabs_needed = set(tabs_needed)
        self.cols_needed = set(cols_needed)
        self.tabs_yielded = set(tabs_yielded)
        self.cols_yielded = set(cols_yielded)
        super().__init__()

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    def all(self) -> T[str, str, str, str]:
        a, b, c, d = tuple(
            map(
                lambda x: ",".join(sorted(x)),
                [
                    self.tabs_needed,
                    self.cols_needed,
                    self.tabs_yielded,
                    self.cols_yielded,
                ],
            )
        )
        return a, b, c, d

    def __str__(self) -> str:
        return pformat(self.__dict__)

    def __bool__(self) -> bool:
        return bool(self.tabs_needed | self.cols_needed | self.tabs_yielded | self.cols_yielded)

    # Public Methods #

    def test(self, other: "Dep") -> bool:
        """Test whether SELF depends on OTHER"""
        return not (
            self.tabs_needed.isdisjoint(other.tabs_yielded)
            and self.cols_needed.isdisjoint(other.cols_yielded)
        )

    @classmethod
    def merge(cls, deps: L["Dep"]) -> "Dep":
        """Combine a list of Deps using UNION"""
        tn, cn, ty, cy = set(), set(), set(), set()  # type: ignore
        for d in deps:
            tn = tn | d.tabs_needed
            cn = cn | d.cols_needed
            ty = ty | d.tabs_yielded
            cy = cy | d.cols_yielded
        return cls(tn, cn, ty, cy)  # type: ignore


################################################################################
class Test:
    """
    Execute a test before running action. If it returns True, the test is
    passed, otherwise it returns an object which is fed into the "message"
    function. This prints a message: "Not Executed (string of object)"
    """

    def __init__(self, test: C[["Generator", Any], bool], message: C[[Any], str]) -> None:
        """
        Initialize test object with callable that takes in a generator and
        returns a bool and a helpful error message.

        Args:
            test (Callable[["Generator",Any]],bool]): Callable that takes in a generator and returns a bool.
            message (Callable[[Any], str]): function that returns a helpful error using the output of the generator.
        """
        self.test = test
        self.message = message

    def __call__(self, t: "Generator", *args: Any) -> Any:
        """Run a test on a generator to see if it's supposed to be executed"""
        output = self.test(t, *args)
        return True if output else self.message(output)


#################
# Example Tests #
#################

onlyTest = Test(
    lambda t, o: (len(o) == 0) or (t.name in o) or any([g in t.tags for g in o]),  # type: ignore
    lambda x: "excluded",
)

xTest = Test(
    lambda t, x: (t.name not in x) and (not any([g in t.tags for g in x])),  # type: ignore
    lambda x: "excluded",
)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING
from typing import Iterator as Iter
from typing import List as L
from typing import Union as U

from infinite import product
from networkx import DiGraph

from dbgen.core.fromclause import Path

# Internal
from dbgen.utils.misc import Base

if TYPE_CHECKING:
    from dbgen.core.schema import Entity, Rel, RelTup
    from dbgen.core.schemaclass import Schema

    Schema, Rel, RelTup, Entity
##############################################################################


class Constraint(Base):
    """
    A way of specifying HOW one arrives at a certain table through JOINing on
    relations. Under normal circumstances, we constrain possible JOIN paths by
    giving a sequence of relations that must be traversed in a certain order.

    This alone would only allow for linear (nonbranching) paths.

    It is possible that a table should be arrived at via multiple
    distinct paths, in which case "branch" should be used.
    """

    def __init__(
        self,
        tab: U[str, "Entity"],
        reqs: L[U["Rel", "RelTup"]] = None,  # constraints on final 'linear' sequence
        xclude: L[U["Rel", "RelTup"]] = None,  # Never use these
        backtrack: bool = False,  # allow the
        branch: L["Constraint"] = None,  # any branching prior to final sequence
    ) -> None:
        self.tab = tab if isinstance(tab, str) else tab.name
        self.reqs = reqs or []
        self.xclude = set(xclude or [])
        self.backtrack = backtrack
        self.branch = set(branch) if branch else set()

        if branch:
            err = "All branches must start from same table %s"
            tabs = {b.tab for b in self.branch}
            assert len(tabs) == 1, err % tabs
        super().__init__()

    def __str__(self) -> str:
        return f"Constraint<{self.tab}>"

    @property
    def branch_obj(self) -> str:
        """Get the branching object, if it exists"""
        return next(iter(self.branch)).tab

    def find(
        self,
        m: "Schema",
        basis: U[str, "Entity", L[U[str, "Entity"]]],
        links: L[U["Rel", "RelTup"]] = None,
        quit: bool = True,
    ) -> None:
        """
        User interface for getting paths
        """
        msg = """Searching for paths from %s to %s\n""" + "%s\n\t" * 4

        # Compute informatino graph in light of the 'links'
        ig = m._info_graph(links or []).reverse()

        # Process basis data
        if not isinstance(basis, list):
            basis_ = [basis]
        else:
            basis_ = basis
        base = [x if isinstance(x, str) else x.name for x in basis_]
        bstr = ",".join(base)

        self._update_reltup(m)

        # underline important things
        l1, l2, l3, l4 = 25, len(bstr), 4, len(self.tab)
        under = " " * l1 + "#" * l2 + " " * l3 + "#" * l4
        constrstr = f"(constraints {self.reqs})" if self.reqs else ""
        branchstr = f"(branch {self.branch})" if self.branch else ""
        linkstr = f"(bypass chickenfeet for {links})" if links else ""

        print(msg % (bstr, self.tab, under, constrstr, branchstr, linkstr))
        for path in self.find_path(ig, base):
            print(repr(path))
            print(path._from().print())
            if input("\nContinue?\n\ty/n -> ").lower() not in ["y", "yes"]:
                break

        if quit:
            exit()

    def find_path(self, schema: DiGraph, basis: L[str]) -> Iter[Path]:
        """Work backwards to find a path to the target that satisfies constraints"""
        # Initialize with empty FK graph, all requirements present
        stack = [(Path(self.tab), list(self.reqs))]
        # Explore possible states in BFS
        while stack:
            # Get current state
            path, reqs = stack.pop(0)
            # print('\n\tnew iteration',path,reqs)

            # Enter the branching logic if we are at branchobj
            if not reqs and self.branch and path.base == self.branch_obj:
                branchiters = [c.find_path(schema, basis) for c in self.branch]
                for branches in product(*branchiters):
                    G = path.copy()
                    for branch in branches:
                        G = G.add_branch(branch)
                    yield G
                print("No more paths")
                return

            if not reqs and path.base in basis:
                yield path  # We're done!
            else:
                # Possible FKs to join on
                for edge in schema[path.base]:
                    seen = set() if self.backtrack else path.all_rels()
                    fks = sorted(set(schema[path.base][edge]["fks"]) - self.xclude - seen)
                    for fk in fks:

                        if reqs:
                            rs = reqs[:-1] if reqs[-1] == fk else list(reqs)
                        else:
                            rs = []
                        stack.append((path.add(fk), rs))

        print("No more paths")

    def _update_reltup(self, m: "Schema") -> "None":
        self.reqs = [m.get_rel(r) for r in self.reqs]
        self.branch = {b.__update_reltup(m) for b in self.branch}

    def __update_reltup(self, m: "Schema") -> "Constraint":
        """Version of _update_reltup that returns a copy"""
        c = self.copy()
        c._update_reltup(m)
        return c



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Common list utilities for flattening, concating, and broadcasting."""
from datetime import datetime
from decimal import Decimal
from typing import Any
from typing import Callable as C
from typing import Dict as D
from typing import Iterable
from typing import List as L
from typing import TypeVar
from typing import Union as U

##############################################################################
A = TypeVar("A")
B = TypeVar("B")
##############################################################
def flatten(lol: L[L[A]]) -> L[A]:
    """Convert list of lists to a single list via concatenation"""
    return [item for sublist in lol for item in sublist]


##############################################################
def nub(seq: L[A], idfun: C = None) -> L[A]:
    """
    Remove duplicates but preserve order.
    """
    if idfun is None:

        def f(x: A) -> A:
            return x

    else:
        f = idfun
    seen = {}  # type: D[A,int]
    result = []  # type: L[A]
    for item in seq:
        marker = f(item)
        if marker in seen:
            continue
        seen[marker] = 1
        result.append(item)
    return result


##############################################################
def merge_dicts(dicts: L[D[A, B]]) -> D[A, B]:
    return {k: v for d in dicts for k, v in d.items()}


##############################################################
def concat_map(f: C[[A], L[B]], args: L[A]) -> L[B]:
    """
    Maps a function over an input.
    We apply the function to every element in the list and concatenate result.
    """
    return flatten([f(arg) for arg in args])


##############################################################
valid_types = (
    int,
    str,
    tuple,
    float,
    list,
    bytes,
    datetime,
    Decimal,
    type(None),
)


def broadcast(args: L[U[L[B], A]]):
    """
    Enforce that all non-length-1 elements have the same length and replicate
    length-1 elements to the largest length list, then zip all the lists
    """

    type_err = "Arg (%s) BAD DATATYPE %s IN NAMESPACE "
    broad_err = "Can't broadcast: maxlen = %d, len a = %d (%s)"
    maxlen = 1  # initialize variable

    for a in args:
        assert isinstance(a, valid_types), type_err % (a, a.__class__)

        if isinstance(a, (list, tuple)):
            if maxlen != 1:  # variable has been set
                # preconditions for broadcasting
                assert len(a) in [1, maxlen], broad_err % (
                    maxlen,
                    len(a),
                    str(a),
                )
            else:
                maxlen = len(a)  # set variable for first (and last) time

    def process_arg(x: U[L[B], A]) -> U[L[B], L[A]]:
        if isinstance(x, (list, tuple)) and len(x) != maxlen:
            return maxlen * list(x)  # broadcast
        elif not isinstance(x, list):
            return maxlen * [x]
        else:
            return x

    # now all args should be lists of the same length
    broadcasted = [process_arg(x) for x in args]
    return list(zip(*broadcasted))


##############################################################
def batch(iterable: list, n: int = 1) -> Iterable:
    """
    returns an iterable that where every iteration
    returns n of items from the original list
    """
    iter_length = len(iterable)
    for ndx in range(0, iter_length, n):
        yield iterable[ndx : min(ndx + n, iter_length)]


##############################################################
def is_iterable(potentially_iterable: Any) -> bool:
    """
    Tests iterability for better error messaging.

    Args:
        potentially_iterable (Any): variable to test for iterability

    Returns:
        bool: potentially_iterable is iterable
    """
    try:
        iter(potentially_iterable)
    except TypeError:
        return False
    return True



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Module for the DBgen Model object"""
from functools import reduce

# External
from typing import Any
from typing import Dict as D
from typing import List as L
from typing import Set as S
from typing import Tuple as T
from typing import Union as U

from hypothesis.strategies import SearchStrategy, just
from networkx import DiGraph

from dbgen.core.expr.sqltypes import SQLType
from dbgen.core.fromclause import Path as JPath
from dbgen.core.func import Env
from dbgen.core.gen import Generator
from dbgen.core.load import Load
from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.core.model.metatable import make_meta
from dbgen.core.model.run import check_patheq, run, validate_name

# Internal
from dbgen.core.model.run_gen import run_gen
from dbgen.core.schema import Attr, AttrTup, Entity, Partition, PathEQ, Rel, RelTup, SuperRel, UserRel, View
from dbgen.core.schemaclass import Schema
from dbgen.utils.exceptions import DBgenInternalError

################################################################################
# Type Synonyms
Stuff = U[
    L[Entity],
    L[Rel],
    L[str],
    L[AttrTup],
    L[RelTup],
    L[Generator],
    L[View],
    L[PathEQ],
    L[T[str, Attr]],
    L[U[Entity, Rel, RelTup, AttrTup, str, Generator, PathEQ, View]],
]
UNIVERSE_TYPE = D[str, T[str, S[str], S[str], D[str, SQLType]]]
##########################################################################################


class Model(Schema):
    """
    Just a named container for objects, relations, and generators

    Also, path equivalencies: which can be checked ad hoc
    """

    def __init__(
        self,
        name: str,
        objlist: L[Entity] = None,
        genlist: L[Generator] = None,
        viewlist: L[View] = None,
        pes: L[PathEQ] = None,
    ) -> None:
        """
        Initialize model with list of objects, generators, views, and path
        equivalency. These lists can be appended to after initizialization
        through the model.add method

        Args:
            name (str): Name of the model used to uniquely identify the model
            objlist (L[Entity], optional): List of Entity objects. Defaults to [].
            genlist (L[Gen], optional): List of Generator objects. Defaults to [].
            viewlist (L[View], optional): List of View objects. Defaults to [].
            pes (L[PathEQ], optional): List of path equivalency objects. Defaults to [].
        """
        self.name = name
        self.objlist = objlist or []
        self.genlist = genlist or []
        self.viewlist = viewlist or []

        self._fks = DiGraph()
        self._fks.add_nodes_from(self.objs)  # nodes are object NAMES

        self.pes = set(pes or [])  # path equivalencies

        for o in self.objs.values():
            for rel in o.fks:
                self._add_relation(rel.to_rel(o.name))
        super(Schema, self).__init__()

    @property
    def gens(self) -> D[str, Generator]:
        return {g.name: g for g in self.genlist}

    @property
    def env(self) -> Env:
        return reduce(lambda first, second: first + second.env, self.genlist, Env())

    def __str__(self) -> str:
        p = "%d objs" % len(self.objs)
        n = self._fks.number_of_edges()
        r = "%d rels" % n if n else ""
        m = "%d gens" % len(self.gens) if self.gens else ""
        e = "%d PathEQs" % len(self.pes) if self.pes else ""

        things = ", ".join(filter(None, [p, r, m, e]))
        return f"Model<{self.name},{things}>"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        """A hypothesis strategy for generating random examples."""
        objs = [
            Entity("a", attrs=[Attr("aa")], fks=[UserRel("ab", "b")]),
            Entity("b", attrs=[Attr("bb")]),
        ]
        gens = [Generator(name="pop_a", transforms=[], loads=[], tags=["io"])]
        return just(cls(name="model", objlist=objs, genlist=gens))

    ######################
    # Externally defined #
    ######################
    run = run

    _validate_name = validate_name
    _run_gen = run_gen
    _make_metatables = make_meta
    _check_patheq = check_patheq

    def run_airflow(self, *args, **kwargs):
        try:
            from dbgen.core.model.run_airflow import run_airflow
        except ImportError as exc:
            print(
                "Import error on model.run_airflow call, apache-airflow is required"
                "for running a model using airflow (This is highly experimental "
                "feature right now)"
            )
            raise exc
        run_airflow(self, *args, **kwargs)

    ##################
    # Public methods #
    ##################
    def test_gen(
        self,
        gen_name: str,
        db: ConnI,
        interact: bool = True,
        limit: int = 5,
    ) -> T[L[D[str, dict]], L[D[str, L[dict]]]]:
        assert gen_name in self.gens, f"Generator {gen_name} not in model:\n{self.gens.keys()}"

        return self.gens[gen_name].test_with_db(
            universe=self._get_universe(), db=db, interact=interact, limit=limit
        )

    def test_transforms(self) -> None:
        """Run all PyBlock tests"""
        for g in self.gens.values():
            for f in g.transforms:
                f.test()

    def check_paths(self, db: ConnI) -> None:
        """Use ASSERT statements to verify one's path equalities are upheld"""
        for pe in self.pes:
            self._check_patheq(pe, db)

    def make_path(self, end: U[str, "Entity"], rels: list = None, name: str = None) -> JPath:
        # Upgrade End
        # Change end into object if it is a string
        if isinstance(end, str):
            upgraded_end = self[end]
        elif isinstance(end, Partition):
            raise NotImplementedError()
        else:
            assert isinstance(end, Entity)
            upgraded_end = end

        # UPGRADE FKS
        def upgrade_rels(fs: list):
            res = []
            for rel_or_list in fs:
                if isinstance(rel_or_list, RelTup):
                    rel_or_list = self.get_rel(rel_or_list)
                if isinstance(rel_or_list, Rel):
                    res.append(
                        SuperRel(
                            rel_or_list.name,
                            rel_or_list.o1,
                            rel_or_list.o2,
                            self[rel_or_list.o2].id_str,
                        )
                    )
                else:
                    res.append(upgrade_rels(rel_or_list))
            return res

        upgraded_fks = upgrade_rels(rels) if rels else []

        return JPath(upgraded_end, upgraded_fks, name=name)

    def get(self, objname: str, partition_val: Any = None) -> Entity:
        """Get an object by name"""
        return self.get_entity(objname, partition_val)

    def rename(self, x: U[Entity, Rel, RelTup, AttrTup, str, Generator], name: str) -> None:
        """Rename an Objects / Relations / Generators / Attr """
        if isinstance(x, (Entity, str)):
            if isinstance(x, str):
                o = self[x]
            else:
                o = x
            self._rename_object(o, name)
        elif isinstance(x, (Rel, RelTup)):
            r = x if isinstance(x, Rel) else self.get_rel(x)
            self._rename_relation(r, name)
        elif isinstance(x, Generator):
            self._rename_gen(x, name)
        elif isinstance(x, AttrTup):
            self._rename_attr(x, name)
        else:
            raise TypeError(f"A {type(x)} ({x}) was passed to rename")

    def add(self, stuff: Stuff) -> None:
        """Add a list containing Objects / Relations / Generators / PathEQs """
        for x in stuff:
            if isinstance(x, (Entity, str)):
                if isinstance(x, str):
                    o = self[x]
                else:
                    o = x
                self._add_object(o)
            elif isinstance(x, (Rel, RelTup)):
                r = x if isinstance(x, Rel) else self.get_rel(x)
                self._add_relation(r)
            elif isinstance(x, Generator):
                self._add_gen(x)
            elif isinstance(x, View):
                self._add_view(x)
            elif isinstance(x, PathEQ):
                self._add_patheq(x)
            elif isinstance(x, tuple) and isinstance(x[1], Attr):
                assert isinstance(x[0], str)
                self._add_attr(x[0], x[1])
            else:
                raise TypeError(f"A {type(x)} ({x}) was passed to add")

    def remove(self, stuff: Stuff) -> None:
        """Remove items given a list of Objects / Relations / Gens / PathEQs"""
        for x in stuff:
            if isinstance(x, (Entity, str)):
                if isinstance(x, str):
                    o = self[x]
                else:
                    o = x
                self._del_object(o)
            elif isinstance(x, (Rel, RelTup)):
                r = x if isinstance(x, Rel) else self.get_rel(x)
                self._del_relation(r)
            elif isinstance(x, Generator):
                self._del_gen(x)
            elif isinstance(x, View):
                self._del_view(x)
            elif isinstance(x, PathEQ):
                self._del_patheq(x)
            else:
                raise TypeError(f"A {type(x)} ({x}) was passed to remove")

    ###################
    # Private Methods #
    ###################

    @classmethod
    def _build_new(cls, name: str) -> "Model":
        """Create a new model (used to generate meta.db)"""
        return cls(name)

    ##################################
    # Adding/removing/renaming in model
    ##################################
    def _rename_gen(self, g: Generator, n: str) -> None:
        assert g == self.gens[g.name]
        g.name = n
        del self.gens[g.name]
        self.gens[n] = g

    def _rename_attr(self, a: AttrTup, n: str) -> None:
        """Replace object with one with a renamed attribute"""
        self.objs[a.obj] = self.objs[a.obj].rename_attr(a.name, n)
        # Make changes in generators?
        # Make changes in PathEQs?

    def _rename_object(self, o: Entity, n: str) -> None:
        """Probably buggy"""
        assert o in self.objs.values(), f"Cannot delete {o}: not found in model"
        oc = o.copy()
        oc.name = n
        del self.objs[o.name]
        self.objs[n] = oc

        for genname, g in self.gens.items():
            self.gens[genname] = g.rename_object(o, n)

        for fk in self._obj_all_fks(o):
            self._del_relation(fk)
            if fk.o1 == o.name:
                fk.o1 = n
            if fk.o2 == o.name:
                fk.o2 = n
            self._add_relation(fk)

    def _rename_relation(self, r: Rel, n: str) -> None:
        raise NotImplementedError

    def _del_gen(self, g: Generator) -> None:
        """Delete a generator"""
        del self.gens[g.name]

    def _del_view(self, v: View) -> None:
        """Delete a view"""
        del self.views[v.name]
        # need to delete generators/PathEQs

    def _del_relation(self, r: Rel) -> None:
        """ Remove from internal FK graph. Need to also remove all Generators that mention it? """
        self._fks[r.o1][r.o2]["fks"].remove(r)
        if not self._fks[r.o1][r.o2]["fks"]:
            self._fks.remove_edge(r.o1, r.o2)

        # Remove any path equivalencies that use this relation
        remove = set()
        for pe in self.pes:
            for p in pe:
                if r.tup() in p.rels:
                    remove.add(pe)
        self.pes -= remove

    def _del_object(self, o: Entity) -> None:
        """Need to also remove all Generators that mention it?"""
        del self.objs[o.name]
        # Remove relations that mention object
        self._fks.remove_node(o.name)

        # Remove generators that mention object
        delgen = []
        for gn, g in self.gens.items():
            gobjs = g.dep(self.objs).tabs_needed | g.dep(self.objs).tabs_yielded
            if o.name in gobjs:
                delgen.append(gn)
        for gn in delgen:
            del self.gens[gn]

        # Remove path equivalencies that mention object
        remove = set()
        for pe in self.pes:
            for p in pe:
                pobjs = set([getattr(p.attr, "obj")] + [r.obj for r in p.rels])
                if o.name in pobjs:
                    remove.add(pe)

        self.pes -= remove

    def _del_patheq(self, peq: PathEQ) -> None:
        self.pes.remove(peq)

    def _add_gen(self, g: Generator) -> None:
        """Add to model"""
        # Validate
        # --------
        if g.name in self.gens:
            raise ValueError(f"Cannot add {g}, name already taken!")
        # Add
        # ----
        for a in g.loads:
            self._validate_load(a)

        self.genlist.append(g)

    ###################
    # Generator related ###
    ###################
    def _validate_load(self, a: Load) -> None:
        """
        It is assumed that an Load provides all identifying data for an
        object, so that either its PK can be selected OR we can insert a new
        row...however this depends on the global model state (identifying
        relations can be added) so a model-level validation is necessary
        The load __init__ already verifies all ID attributes are present
        """
        # Check all identifying relationships are covered
        if not a.pk:  # don't have a PK, so need identifying info
            for fk in self[a.obj].id_fks():
                err = "%s missing identifying relation %s"
                assert fk in a.fks, err % (a, fk)

        # Recursively validate sub-loads
        for act in a.fks.values():
            self._validate_load(act)

    def ordered_gens(self) -> L[Generator]:
        """ Determine execution order of generators """

        # Check for cycles:
        from networkx.algorithms import simple_cycles

        from dbgen.utils.graphs import topsort_with_dict

        sc = list(simple_cycles(self._gen_graph()))

        if sc:
            small = min(sc, key=len)
            print("\nGenerator cycle found! Smallest cycle:")
            for gname in small:
                g = self.gens[gname]
                print("\n################\n", g.name)
                for k, v in vars(g.dep(self.objs)).items():
                    print(k, v)
            raise DBgenInternalError("Found Cycle")

        # Use topological sort to order
        return list(topsort_with_dict(self._gen_graph(), self.gens))

    def _gen_graph(self) -> "DiGraph":
        """ Make a graph out of generator dependencies."""
        from networkx import DiGraph

        G = DiGraph()

        ddict = {a: g.dep(self.objs) for a, g in self.gens.items()}
        G.add_nodes_from(list(self.gens.keys()))
        for a1 in self.gens.keys():
            d1 = ddict[a1]
            for a2 in self.gens.keys():
                d2 = ddict[a2]
                if a1 != a2 and d1.test(d2):
                    G.add_edge(a2, a1)
        return G

    def _todo(self) -> S[str]:
        """ All attributes that do not yet have a generator populating them """
        allattr = set()
        for o in self.objs.values():
            allattr.update([o.name + "." + a for a in o.attrnames()])
        for r in self._rels():
            allattr.add(r.o1 + "." + r.name)
        alldone = set()
        for g in self.gens.values():
            alldone.update(g.dep(self.objs).cols_yielded)
        return allattr - alldone

    def _get_universe(self) -> UNIVERSE_TYPE:
        return {
            oname: (
                o.id_str,
                set(o.ids()),
                set(o.id_fks()),
                {key: val.dtype for key, val in o.attrdict.items()},
            )
            for oname, o in self.objs.items()
        }



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Initialize dbgen"""
import logging
from pathlib import Path
from typing import Optional

from dbgen.core.misc import ConnectInfo
from dbgen.utils.config import config
from dbgen.utils.log import setup_logger

logger = logging.getLogger("dbgen.settings")
# Initalize variables
DB_CONN_STR: Optional[str] = None
METADB_CONN_STR: Optional[str] = None
DB_SCHEMA: Optional[str] = None
METADB_SCHEMA: Optional[str] = None
CONN: Optional[ConnectInfo] = None
META_CONN: Optional[ConnectInfo] = None


def configure_vars():
    global DB_CONN_STR
    global METADB_CONN_STR
    global DB_SCHEMA
    global METADB_SCHEMA
    DB_CONN_STR = config.get("core", "DB_CONN_STR")
    METADB_CONN_STR = config.get("core", "METADB_CONN_STR", fallback=None)
    DB_SCHEMA = config.get("core", "DB_SCHEMA", fallback=None)
    METADB_SCHEMA = config.get("core", "METADB_SCHEMA", fallback=None)


def configure_connections():
    """
    Cycles through database connection variables until first on set is found.

    Order is db_connection_str
    !TODO! Implement airflow_conn_id
    """
    global CONN
    global META_CONN
    # Main Database
    if DB_CONN_STR:
        CONN = ConnectInfo.from_dsn(DB_CONN_STR, DB_SCHEMA)
    else:
        raise ValueError(
            "No Connection string at any of the following locations:"
            + "\n".join(["DBGEN__CORE__DB_CONN_STR", "config(core.DB_CONN_STR)"])
        )

    if METADB_CONN_STR:
        META_CONN = ConnectInfo.from_dsn(METADB_CONN_STR, METADB_SCHEMA)
    else:
        logger.warning(
            f"Did not find a envvar or config value for meta_db conn using CONN_STR with schema {CONN.schema}_log"
        )
        META_CONN = CONN.copy()
        if METADB_SCHEMA:
            META_CONN.schema = METADB_SCHEMA
        else:
            META_CONN.schema = META_CONN.schema + "_log"
        assert META_CONN != CONN


def configure_logging():
    log_path = config.get("logging", "log_path")
    log_level = config.get("logging", "log_level")
    write_logs = config.getboolean("logging", "write_logs")
    log_format = config.get("logging", "log_format")
    log_to_stdout = config.getboolean("logging", "log_to_stdout")
    setup_logger(
        "dbgen",
        log_level,
        write_logs=write_logs,
        log_path=Path(log_path),
        log_format=log_format,
        log_to_stdout=log_to_stdout,
    )


def initialize():
    configure_logging()
    configure_vars()
    configure_connections()



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External Modules
from typing import TYPE_CHECKING, Any
from typing import List as L
from typing import Optional

from tqdm import tqdm

# Internal Modules
from dbgen.core.expr.sqltypes import Boolean, Decimal, Int, Text, Timestamp, Varchar
from dbgen.core.schema import Attr, Entity
from dbgen.core.schema import UserRel as Rel
from dbgen.utils.sql import mkInsCmd, mkSelectCmd, sqlexecute, sqlselect
from dbgen.utils.str_utils import hash_

# Internal Modules
if TYPE_CHECKING:
    from dbgen.core.gen import Generator
    from dbgen.core.misc import ConnectInfo as ConnI
    from dbgen.core.model.model import Model

    ConnI, Model, Generator


#############################################################################
def safex(conn: Any, q: str, binds: list) -> None:
    sqlexecute(conn, q, binds)


###########
# Constants
##########
objs = [
    Entity(
        "connection",
        "Info required to connect to a PostGres DB",
        attrs=[
            Attr("hostname", Varchar(), identifying=True),
            Attr("user", Varchar(), identifying=True),
            Attr("port", identifying=True),
            Attr("db", Varchar(), identifying=True),
        ],
    ),
    Entity(
        "temp",
        desc="Temporary table that is populated and truncated after checking for repeat values",
        attrs=[
            Attr(
                "ind",
                identifying=True,
                desc="Index to a list of query-generated inputs",
            )
        ],
    ),
    Entity(
        "object",
        "All static info about a given class of entities being modeled",
        attrs=[
            Attr("name", Varchar(), identifying=True),
            Attr("description", Text()),
        ],
    ),
    Entity(
        "attr",
        "Property of an object",
        attrs=[
            Attr("name", Varchar(), identifying=True),
            Attr("dtype", Varchar()),
            Attr("description", Text()),
            Attr("defaultval", Text()),
        ],
        fks=[Rel("object", identifying=True)],
    ),
    Entity(
        "view",
        "SQL view",
        attrs=[
            Attr("name", Varchar(), identifying=True),
            Attr("query", Text("long")),
        ],
    ),
    Entity(
        "func",
        "Python functions that get used during generation of Objects/Attributes",
        attrs=[
            Attr("source", Text(), identifying=True),
            Attr("name", Varchar()),
        ],
    ),
    Entity(
        "gen",
        "Method for generating concrete data",
        attrs=[
            Attr("name", Varchar(), identifying=True),
            Attr("description", Text()),
            Attr("gen_json", Text()),
        ],
    ),
    Entity(
        "pyblock",
        "decorated python function",
        attrs=[],
        fks=[Rel("gen", identifying=True), Rel("func")],
    ),
    Entity(
        "const",
        "A constant injected into the namespace of an generator",
        attrs=[
            Attr("dtype", Varchar(), identifying=True),
            Attr("val", Text(), identifying=True),
        ],
    ),
    Entity(
        "arg",
        "How a PyBlock refers to a namespace",
        attrs=[
            Attr("ind", Int(), identifying=True),
            Attr("keyname", Varchar()),
            Attr("name", Varchar()),
        ],
        fks=[Rel("const")],
    ),
    Entity(
        "run",
        "Each time DbGen is run, a new Run instance is created",
        attrs=[
            Attr("starttime", Timestamp(), default="CURRENT_TIMESTAMP"),
            Attr("start", Varchar()),
            Attr("until_", Varchar()),
            Attr("delta", Decimal(), desc="Runtime in minutes"),
            Attr("status", Varchar(), desc="Status of run"),
            Attr("errs", Int()),
            Attr("retry", Boolean()),
            Attr("nuke", Varchar()),
            Attr("onlyrun", Varchar()),
            Attr("exclude", Varchar()),
        ],
        fks=[Rel("connection")],
    ),
    Entity(
        "gens",
        "A list of Generator instances associated with a given run",
        attrs=[
            Attr("name", Varchar()),
            Attr("status", Varchar()),
            Attr("runtime", Decimal()),
            Attr("n_inputs", Int()),
            Attr("rate", Decimal()),
            Attr("error", Text("long")),
            Attr("description", Text()),
            Attr("query", Text()),
            Attr("ind", Int()),
            Attr("tabdep", Text()),
            Attr("coldep", Text()),
            Attr("newtab", Text()),
            Attr("newcol", Text()),
            Attr("basis", Varchar()),
        ],
        fks=[Rel("run", identifying=True), Rel("gen", identifying=True)],
    ),
    Entity(
        "objs",
        "A list of Object instances associated with a given run",
        fks=[Rel("object", identifying=True), Rel("run", identifying=True)],
    ),
    Entity(
        "views",
        "List of View instances associated with a given run",
        fks=[Rel("view", identifying=True), Rel("run", identifying=True)],
    ),
    Entity(
        "repeats",
        "A record of which inputs a given Load has already seen",
        fks=[Rel("gen", identifying=True), Rel("run")],
    ),
]

# Views
create_curr_run = """
CREATE OR REPLACE VIEW curr_run AS
    SELECT name,status,runtime,n_inputs,rate,error,query,description,
           tabdep,coldep,newtab,newcol,basis
    FROM gens
    WHERE gens.run = (SELECT max(run.run_id) FROM run)
    ORDER BY gens.ind
"""
create_all_run = """
CREATE OR REPLACE VIEW all_run AS
select
    run.run_id,
    "name",
    gens.status,
    runtime,
    n_inputs,
    rate,
    error,
    description,
    query,
    ind
from
    gens
join run on
    gens.run = run.run_id
where run_id = (select max(run_id) from run) or run.status in ('running','initialized')
order by
    run_id desc,
    gens.ind
"""

#############################################################################
# Main function


def make_meta(
    self: "Model",
    mconn: "ConnI",
    conn: "ConnI",
    nuke: bool,
    retry: bool,
    only: L[str],
    xclude: L[str],
    start: Optional[str],
    until: Optional[str],
    bar: bool,
) -> int:
    """
    Initialize metatables
    """

    NUKE_META = True  # whether or not to erase metatable data if nuking DB
    meta = self._build_new("meta")
    meta.add(objs)

    ################################################################################

    if nuke and NUKE_META:
        mconn.drop()
        mconn.create()
        gmcxn = mconn.connect()
    else:
        try:
            gmcxn = mconn.connect()
        except Exception:
            raise Exception("When making DB for first time, run with --nuke=True")

    # Create metatables if they don't exist
    # --------------------------------------
    for t in meta.objs.values():
        for sql in t.create():

            sqlexecute(gmcxn, sql)

    for r in meta._rels():
        sqlexecute(gmcxn, meta._create_fk(r))

    sqlexecute(gmcxn, create_curr_run)
    sqlexecute(gmcxn, create_all_run)

    # Create new run instance
    # -------------------------
    run_id = sqlselect(gmcxn, "SELECT MAX(run_id)+1 FROM run")[0][0] or 1

    # Insert connection (if it dosn't exist already)
    # ----------------------------------------------
    cxn_cols = ["connection_id", "hostname", "user", "port", "db"]
    cxn_args_ = [conn.host, conn.user, conn.port, conn.db]
    cxn_args = [hash_("$".join(map(str, cxn_args_)))] + cxn_args_  # type: ignore
    cxn_sql = mkInsCmd("connection", cxn_cols)
    sqlexecute(gmcxn, cxn_sql, cxn_args)

    # Get current connection ID
    # ----------------------------------------------
    get_cxn = mkSelectCmd("connection", ["connection_id"], ["connection_id"])
    cxn_id = sqlselect(gmcxn, get_cxn, cxn_args[:1])[0][0]

    # Insert top level information about current run
    # ----------------------------------------------
    run_cols = [
        "run_id",
        "status",
        "retry",
        "onlyrun",
        "exclude",
        "nuke",
        "connection",
        "start",
        "until_",
    ]
    run_status = "initializing"
    run_args = [
        run_id,
        run_status,
        retry,
        " ".join(only),
        " ".join(xclude),
        nuke,
        cxn_id,
        start,
        until,
    ]

    fmt_args = [",".join(run_cols), ",".join(["%s"] * len(run_args))]
    run_sql = "INSERT INTO run ({}) VALUES ({})".format(*fmt_args)

    sqlexecute(gmcxn, run_sql, run_args)

    # Insert info about current DBG if it doesn't exist
    # ----------------------------------------------
    od = "Inserting Objects into MetaDB"
    ad = "Inserting Loads into MetaDB"
    vd = "Inserting Views into MetaDB"
    oq = mkInsCmd("objs", ["run", "object", "objs_id"])
    vq = mkInsCmd("views", ["run", "view", "views_id"])
    aq = mkInsCmd(
        "gens",
        [
            "run",
            "gen",
            "name",
            "status",
            "description",
            "query",
            "ind",
            "tabdep",
            "coldep",
            "newtab",
            "newcol",
            "basis",
            "gens_id",
        ],
    )
    tqargs = dict(leave=False, disable=not bar)
    for o in tqdm(self.objs.values(), desc=od, **tqargs):
        safex(gmcxn, oq, [run_id, o.add(gmcxn), hash_(str(run_id) + str(o.hash))])

    for vn, v in tqdm(self.views.items(), desc=vd, **tqargs):
        safex(gmcxn, vq, [run_id, v.add(gmcxn), hash_(str(run_id) + str(v.hash))])

    for i, u in enumerate(tqdm(self.ordered_gens(), desc=ad, **tqargs)):
        # print(i,u)
        v = u.query
        q = v.showQ() if v else ""
        td, cd, nt, nc = u.dep(self.objs).all()
        b = ",".join(v.basis) if v else ""

        safex(
            gmcxn,
            aq,
            [
                run_id,
                u.add(gmcxn),
                u.name,
                "initialized",
                u.desc,
                q,
                i,
                td,
                cd,
                nt,
                nc,
                b,
                hash_(str(run_id) + str(u.hash)),
            ],
        )
    return run_id



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Internal Modules
from dbgen import SUM, Generator, Model, Query

################################################################################


def analysis(mod: Model) -> None:

    # Extract tables
    tabs = ["employee", "department"]

    # Extract Step
    Emp, Dept = map(mod.get, tabs)
    emp_path = mod.make_path("employee", [Emp.r("department")])
    exprs = {"tot_salary": SUM(Emp["salary"](emp_path)), "emp_id": Dept.id()}
    query = Query(exprs=exprs, aggcols=[Dept.id()], basis=["department"])
    # Load Step
    dept_load = Dept(department=query["emp_id"], total_salary=query["tot_salary"])

    get_tot_salary = Generator(
        name="get_total_salary",
        desc="Salary",
        transforms=[],
        tags=["analysis"],
        loads=[dept_load],
        query=query,
    )

    ################################################################################
    gens = [get_tot_salary]
    mod.add(gens)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# External imports
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import List as L
from typing import Tuple as T

if TYPE_CHECKING:
    from dbgen.core.misc import ConnectInfo as ConnI
    from dbgen.core.model.model import Model
    from dbgen.utils.sql import Connection as Conn

    Model
    ConnI
    Conn

from airflow.hooks.postgres_hook import PostgresHook
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

# Internal Imports
from dbgen.core.gen import Generator
from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.core.model.run_gen import run_gen
from dbgen.utils.exceptions import DBgenGeneratorError
from dbgen.utils.sql import mkSelectCmd, sqlselect


class GenOperator(BaseOperator):
    """A custom airflow operator for the generator object"""

    template_fields = ()
    ui_color = "#b3cde0"

    @apply_defaults
    def __init__(
        self,
        objs: D[str, T[L[str], L[int], L[int]]],
        gen_name: str,
        gen_hash: int,
        run_id: int,
        db_conn_id: str,
        mdb_conn_id: str,
        retry: bool = False,
        serial: bool = False,
        bar: bool = False,
        user_batch_size: int = None,
        **kwargs: dict,
    ) -> None:
        super().__init__(**kwargs)  # type: ignore
        # Initialize variables
        self.objs = objs
        self.gen_name = gen_name
        self.gen_hash = gen_hash
        self.run_id = run_id
        self.retry = retry
        self.serial = serial
        self.bar = bar
        self.user_batch_size = user_batch_size
        self.db_conn_id = db_conn_id
        self.mdb_conn_id = mdb_conn_id

    def _get_gen(self, mcxn: "Conn") -> Generator:
        get_a = mkSelectCmd("gen", ["gen_json"], ["gen_id"])
        gen_json = sqlselect(mcxn, get_a, [self.gen_hash])[0][0]
        gen = Generator.fromJSON(gen_json)  # type: ignore
        return gen  # type: ignore

    def execute(self, context: Any) -> None:

        # Get the db connections
        gcxn = PostgresHook(self.db_conn_id).get_conn()
        mgcxn = PostgresHook(self.mdb_conn_id).get_conn()

        gcxn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        mgcxn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        conn_info = ConnI.from_postgres_hook(PostgresHook.get_connection(self.db_conn_id))
        mconn_info = ConnI.from_postgres_hook(PostgresHook.get_connection(self.mdb_conn_id))

        gen = self._get_gen(mgcxn)
        assert gen.hash == self.gen_hash, "Serialization error, The gen hash doesn't has changed!"

        run_gen_args = dict(
            self=None,
            objs=self.objs,
            gen=gen,
            run_id=self.run_id,
            gcxn=gcxn,
            gmcxn=mgcxn,
            conn_info=conn_info,
            mconn_info=mconn_info,
            retry=self.retry,
            serial=self.serial,
            bar=self.bar,
            user_batch_size=self.user_batch_size,
            gen_hash=self.gen_hash,
        )

        err = run_gen(**run_gen_args)  # type: ignore
        if err:
            raise DBgenGeneratorError(f"{self.gen_name} failed")



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod
from copy import deepcopy
from importlib import import_module
from inspect import getfullargspec
from json import dumps, loads
from string import ascii_lowercase
from typing import Any
from typing import Dict as D
from typing import List as L
from typing import Type, TypeVar
from typing import Union as U

from hypothesis import infer
from hypothesis.strategies import SearchStrategy, booleans, builds, integers, none, one_of, text

from dbgen.utils.str_utils import hash_

NoneType = type(None)

##############################################################################

T = TypeVar("T")


def identity(x: T) -> T:
    return x


def kwargs(x: Any) -> L[str]:
    return sorted(getfullargspec(type(x))[0][1:])


anystrat = one_of(text(), booleans(), text(), integers(), none())
nonempty = text(min_size=1)
nonempty_limited = text(min_size=1, max_size=3)
letters = text(min_size=1, alphabet=ascii_lowercase)


def build(typ: Type) -> SearchStrategy:
    """Unfortunately, hypothesis cannot automatically ignore default kwargs."""
    args, _, _, _, _, _, annotations = getfullargspec(typ)
    # use default kwarg value if type is Any
    kwargs = {k: infer for k in args[1:] if annotations[k] != Any}
    return builds(typ, **kwargs)


simple = ["int", "str", "float", "NoneType", "bool"]
complex = ["tuple", "list", "set", "dict"]


def to_dict(x: Any, id_only: bool = False) -> U[L, int, float, str, D[str, Any], NoneType]:
    """Create JSON serializable structure for arbitrary Python/DbGen type."""
    module, ptype = type(x).__module__, type(x).__name__
    metadata = dict(_pytype=module + "." + ptype)  # type: D[str, Any]
    if module == "builtins" and ptype in simple:
        return x
    elif module == "builtins" and ptype in complex:
        if ptype == "dict":
            assert all([isinstance(k, str) for k in x.keys()]), x
            return {k: to_dict(v, id_only) for k, v in x.items() if not (k == "_uid" and id_only)}
        elif ptype == "list":
            return [to_dict(xx, id_only) for xx in x]
        elif ptype == "tuple":
            return dict(**metadata, _value=[to_dict(xx, id_only) for xx in x])
        elif ptype == "set":
            return dict(**metadata, _value=[to_dict(xx, id_only) for xx in sorted(x)])
        else:
            raise TypeError(x)
    else:
        assert hasattr(x, "__dict__"), metadata
        data = {
            k: to_dict(v, id_only)
            for k, v in sorted(vars(x).items())
            if (k in kwargs(x)) or (not id_only and k[0] != "_")
        }
        if not id_only:
            hashdict = {
                **metadata,
                **{k: to_dict(v, True) for k, v in sorted(vars(x).items()) if (k in kwargs(x))},
            }
            metadata["_uid"] = hash_(hashdict)
        return {**metadata, **data}


def from_dict(x: Any) -> Any:
    """Create a python/DbGen type from a JSON serializable structure."""
    if isinstance(x, dict):
        ptype = x.get("_pytype", "")
        if "dbgen" in ptype:
            mod, cname = ".".join(ptype.split(".")[:-1]), ptype.split(".")[-1]
            constructor = getattr(import_module(mod), cname)
            return constructor(
                **{k: from_dict(v) for k, v in x.items() if k in getfullargspec(constructor)[0][1:]}
            )
        elif ptype in ("builtins/tuple", "builtins.tuple"):
            return tuple([from_dict(xx) for xx in x["_value"]])  # data-level tuple
        elif ptype in ("builtins/set", "builtins.set"):
            return {from_dict(xx) for xx in x["_value"]}  # data-level tuple
        else:
            assert "_ptype" not in x
            return {k: from_dict(v) for k, v in x.items()}  # data-level dict
    elif isinstance(x, (int, float, list, type(None), str)):
        if isinstance(x, list):
            return [from_dict(xx) for xx in x]
        else:
            return x
    else:
        raise TypeError(x)


class Base(metaclass=ABCMeta):
    """Common methods shared by many DbGen objects."""

    def __init__(self) -> None:
        fields = set(vars(self))
        args = set(kwargs(self))
        missing = args - fields
        assert not missing, f"Need to store args {missing} of {self}"
        assert not any([a[0] == "_" for a in args]), args

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        """
        Maybe the below should be preferred? Try it out, sometime!
        return type(self) == type(other) and vars(self) == vars(other)
        """
        if type(other) == type(self):
            return hash(self) == hash(other)
        else:
            args = [self, type(self), other, type(other)]
            err = "Equality type error \n{} \n({}) \n\n{} \n({})"
            raise ValueError(err.format(*args))

    def copy(self: T) -> T:
        return deepcopy(self)

    def toJSON(self, id_only=False) -> str:
        return dumps(to_dict(self, id_only), indent=4, sort_keys=True)

    @staticmethod
    def fromJSON(s: str) -> "Base":
        val = from_dict(loads(s))
        assert isinstance(val, Base)
        return val

    def __hash__(self) -> int:
        return int(self.hash)

    @property
    def hash(self) -> str:
        return hash_(to_dict(self, id_only=True))

    @classmethod
    def canonical_name(cls) -> str:
        return cls.__module__ + "." + cls.__qualname__



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Extenral
from abc import ABCMeta, abstractmethod
from traceback import format_exc
from typing import Any
from typing import Callable as C
from typing import Dict as D
from typing import List as L
from typing import Sequence
from typing import Tuple as T
from typing import Union as U

from hypothesis.strategies import SearchStrategy, builds, just, one_of

# Internal
from dbgen.core.func import Env, Func
from dbgen.core.misc import ConnectInfo as Conn
from dbgen.utils.exceptions import (
    DBgenExternalError,
    DBgenInvalidArgument,
    DBgenMissingInfo,
    DBgenSkipException,
)
from dbgen.utils.lists import is_iterable
from dbgen.utils.misc import Base, anystrat
from dbgen.utils.sql import mkInsCmd, sqlexecute

"""
The "T" of ETL: things that are like functions.

A generalization of a function is a computational graph (where nodes are functions)

These nodes (PyBlocks) have Args which refer to other PyBlocks (or a query)
"""
####################################################################################
class ArgLike(Base, metaclass=ABCMeta):
    @abstractmethod
    def arg_get(self, dic: dict) -> Any:
        raise NotImplementedError

    @abstractmethod
    def make_src(self, meta: bool = False) -> str:
        raise NotImplementedError

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return one_of(Arg._strat(), Const._strat())


class Arg(ArgLike):
    """
    How a function refers to a namespace
    """

    def __init__(self, key: str, name: str) -> None:
        self.key = str(key)
        self.name = name.lower()

    def __str__(self) -> str:
        return f"Arg({str(self.key)[:4]}...,{self.name})"

    # Public methods #
    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)

    def arg_get(self, dic: dict) -> Any:
        """
        Common interface for Const and Arg to get values out of namespace
        """
        try:
            val = dic[self.key][self.name]
            return val
        except KeyError:
            if self.key not in dic:
                raise DBgenMissingInfo(
                    f"could not find hash, looking for {self.name} at this hash {self.key}"
                )
            else:
                err = "could not find '%s' in %s "
                raise DBgenMissingInfo(err % (self.name, list(dic[self.key].keys())))

    def add(self, cxn: Conn, act: int, block: int) -> None:
        q = mkInsCmd("_arg", ["gen_id", "block_id", "hashkey", "name"])
        sqlexecute(cxn, q, [act, block, self.key, self.name])

    def make_src(self, meta: bool = False) -> str:
        key = repr(self.key) if isinstance(self.key, str) else self.key
        if meta:
            return f"Arg({key},'{self.name}')"
        else:
            return f'namespace[{key}]["{self.name}"]'


class Const(ArgLike):
    def __init__(self, val: Any) -> None:
        if callable(val):
            val = Func.from_callable(val)
        self.val = val

    def __str__(self) -> str:
        return f"Const<{self.val}>"

    def arg_get(self, _: dict) -> Any:
        return self.val

    def make_src(self, meta: bool = False) -> str:
        if meta:
            return "Const(%s)" + repr(self.val)
        else:
            return repr(self.val)

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls, val=anystrat)


class PyBlock(Base):
    """
    A computational block that executes Python code
    """

    def __init__(
        self,
        func: U[C, Func],
        env: Env = Env(),
        args: Sequence[ArgLike] = [],
        outnames: L[str] = None,
        tests: L[T[tuple, Any]] = [],
    ) -> None:

        # Store fields
        self.func = Func.from_callable(func, env=env)
        self.env = env
        self.args = args
        self.outnames = [o.lower() for o in outnames or ["out"]]
        self.tests = tests
        # Validate inputs for errors
        self.validate_inputs()

        super().__init__()

    @classmethod
    def _strat(cls) -> SearchStrategy:
        func = lambda x: x + 1
        return just(
            cls(
                func=func,
                args=[Const(1)],
                outnames=["x"],
            )
        )

    def __str__(self) -> str:
        return "PyBlock<%d args>" % (len(self.args))

    def __lt__(self, other: "PyBlock") -> bool:
        return self.hash < other.hash

    def __call__(self, curr_dict: D[str, D[str, Any]]) -> D[str, Any]:
        """
        Take a TempFunc's function and wrap it such that it can accept a namespace
            dictionary. The information about how the function interacts with the
            namespace is contained in the TempFunc's args.
        """
        try:
            inputvars = [arg.arg_get(curr_dict) for arg in self.args]
        except (TypeError, IndexError) as e:
            print(e)
            raise ValueError()
        except AttributeError:
            invalid_args = [getattr(arg, "arg_get", None) is None for arg in self.args]
            missing_args = filter(lambda x: invalid_args[x], range(len(invalid_args)))
            raise DBgenMissingInfo(
                f"Argument(s) {' ,'.join(map(str,missing_args))} to {self.func.name} don't have arg_get attribute:\n Did you forget to wrap a Const around a PyBlock Arguement?"
            )

        try:
            output = self.func(*inputvars)
            if isinstance(output, tuple):
                l1, l2 = len(output), len(self.outnames)
                assert l1 == l2, "Expected %d outputs from %s, got %d" % (
                    l2,
                    self.func.name,
                    l1,
                )
                return {o: val for val, o in zip(output, self.outnames)}
            else:
                assert len(self.outnames) == 1
                return {self.outnames[0]: output}
        except DBgenSkipException:
            raise
        except Exception:
            msg = f"\tApplying func {self.func.name} in tempfunc:\n\t"
            raise DBgenExternalError(msg + format_exc())

    def __getitem__(self, key: str) -> Arg:
        err = "%s not found in %s"
        assert key.lower() in self.outnames, err % (key, self.outnames)
        return Arg(self.hash, key.lower())

    def _constargs(self) -> L[Func]:
        """all callable constant args"""
        return [c.val for c in self.args if isinstance(c, Const) and isinstance(c.val, Func)]

    def validate_inputs(self) -> None:
        # Check for iterability
        if not is_iterable(self.outnames):
            raise DBgenInvalidArgument(
                f"Func Name: {self.func.name}.\nOutnames is not iterable. Please wrap in singleton list if single argument"
            )
        elif not is_iterable(self.args):
            raise DBgenInvalidArgument(
                f"Func Name: {self.func.name}.\nArgs is not iterable. Please wrap in singleton list if single argument"
            )
        # Check for duplicate outnames
        if self.outnames:
            dups = [o for o in self.outnames if self.outnames.count(o) > 1]
            if dups:
                err = f"No duplicates in outnames for func {self.func.name}: {dups}"
                raise DBgenInvalidArgument(err)
        invalid_args = [i for i, arg in enumerate(self.args) if not isinstance(arg, ArgLike)]
        if invalid_args:
            raise DBgenMissingInfo(
                f"Argument(s) {' ,'.join(map(str,invalid_args))} to {self.func.name} are not ArgLike:\n Did you forget to wrap a Const around a PyBlock Arguement?"
            )

    def add(self, cxn: Conn, a_id: int, b_id: int) -> None:
        """Add to metaDB"""
        cols = ["gen_id", "py_block_id", "func_id", "outnames"]
        f_id = self.func.add(cxn)
        q = mkInsCmd("_py_block", cols)
        sqlexecute(cxn, q, [a_id, b_id, f_id, ",".join(self.outnames)])

    def test(self) -> None:
        """Throw error if any tests are failed"""
        for args, target in self.tests:
            assert self.func(*args) == target

    def make_src(self) -> str:
        # load the pyblock template
        from dbgen.templates import jinja_env

        pb_template = jinja_env.get_template("pyblock.py.jinja")

        # Prepare the template args
        args = [a.make_src() for a in self.args]
        name = self.func.name.replace("<lambda>", "func")
        src = ("func=" if "<lambda>" in self.func.name else "") + self.func.src.replace("\t", "   ")

        # Set the template args
        template_kwargs = dict(src=src, outnames=self.outnames, args=args, name=name)
        rendered_template = pb_template.render(**template_kwargs)
        tabified_template = rendered_template.replace("\n", "\n    ")
        return tabified_template



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import defaultdict
from csv import DictReader
from json import load
from os.path import join
from re import findall
from sqlite3 import connect
from typing import Any
from typing import Dict as D
from typing import List as L
from typing import Tuple as T

from dbgen.example.constants import ROOT

AnyD = D[str, Any]
################################################################################
chemical_symbols = [
    "X",
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
]

# Parse ssn.json
def parse_ssn(file_name: str) -> T[L[str], L[str], L[int]]:
    """Expects a JSON file with <<"firstName lastName" : SSN#>> entries"""
    pth = join(ROOT, file_name)
    with open(pth) as f:
        data = load(f)
    firstnames, lastnames = map(list, zip(*[d.split() for d in data.keys()]))
    ssns = list(map(int, data.values()))
    return firstnames, lastnames, ssns  # type: ignore


# Parse procedures.csv


def parse_proc_csv(
    file_name: str,
) -> T[T[int], T[int], T[str], T[int], T[int], T[str], T[str], T[str]]:
    """Expects CSV: Sample,Step,Procedure,Timestamp,Researcher,Notes"""
    pth = join(ROOT, file_name)
    output = defaultdict(list)  # type: AnyD
    with open(pth) as f:
        reader = DictReader(f)
        for row in reader:
            for k, v in row.items():
                output[k].append(v)

    samp_, step_, proc_, time, ssn_, note = map(list, output.values())
    dtype, name = ["str"] * len(samp_), ["notes"] * len(samp_)
    samp, step, ssn = [list(map(int, x)) for x in [samp_, step_, ssn_]]  # type: ignore
    proc = [x.strip() for x in proc_]  # type: ignore
    return samp, step, proc, time, ssn, note, dtype, name  # type: ignore


# Parse experiment.json
def parse_expt(file_name: str) -> T[L[int], L[str], L[float], L[str]]:
    """Parse JSON file with experiments containing anode/cathode/capacity/date"""
    pth = join(ROOT, file_name)
    expt_ids, dates, capacities, solvents = [], [], [], []
    with open(pth) as f:
        data = load(f).items()

    for expt_id, expt in data:
        expt_ids.append(expt_id)
        dates.append(expt.get("date"))
        solvents.append(expt.get("solvent"))
        capacities.append(expt.get("capacitance"))

    return expt_ids, dates, capacities, solvents  # type: ignore


# Parse experiment.json
def get_electrode(file_name: str, x: str) -> T[L[int], L[int], L[str]]:
    """Get either anode or cathode information about all battery experiments"""
    pth = join(ROOT, file_name)
    electrode = x.lower()
    assert electrode in ["anode", "cathode"]

    ids, expt_ids, compositions = [], [], []  # type: ignore

    with open(pth) as f:
        data = load(f).items()

    for expt_id, expt in data:
        expt_ids.append(expt_id)
        ids.append(expt[electrode]["id"])
        compositions.append(expt[electrode]["composition"])

    return ids, expt_ids, compositions


# Parse procedure.db
def parse_sqlite(file_name: str) -> T[L[int]]:
    """Get sample history data stored in a relational db"""
    pth = join(ROOT, file_name)
    db = connect(pth).cursor()
    cols = ",".join(["sample", "step", "procedure", "firstname", "lastname", "ssn"])
    output = db.execute(f"SELECT {cols} FROM test JOIN scientist USING (scientist_id)")
    return tuple(map(list, zip(*output)))  # type: ignore


# Extract element info from a chemical formula string
def parse_formula(formula: str) -> T[L[int], L[str], L[float]]:
    """Parses simple chemical formulas like As2Ga3 and H2O"""
    matches = findall(r"([A-Z][a-z]?)(\d*)", formula)
    symbs, nums = map(list, zip(*matches))  # type: T[L[Any],L[Any]]
    atomic_nums = list(map(chemical_symbols.index, symbs))
    nums = [int(n) if n else 1 for n in nums]
    fracs = [n / sum(nums) for n in nums]
    return atomic_nums, symbs, fracs



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
The Query class, as well as Ref (used to indirectly refer to an object in a
query without knowing the exact join path)

Furthermore some Model methods that are highly related to queries are defined.
"""
# External
from typing import Any
from typing import Callable as C
from typing import List as L
from typing import Mapping as M
from typing import Sequence
from typing import Set as S
from typing import Union as U

from hypothesis.strategies import SearchStrategy, builds, dictionaries, lists

from dbgen.core.expr.expr import PK, Expr, true
from dbgen.core.expr.pathattr import PathAttr, expr_attrs

# Internal
from dbgen.core.fromclause import From
from dbgen.core.funclike import Arg
from dbgen.core.misc import ConnectInfo as ConnI
from dbgen.core.schema import Entity, RelTup
from dbgen.utils.lists import flatten, nub
from dbgen.utils.misc import nonempty
from dbgen.utils.sql import select_dict

Fn = C[[Any], str]  # type shortcut

################################################################################
class Query(Expr):
    """
    Specification of a query, which can only be realized in the context of a model

    exprs - things you SELECT for, keys in dict are the aliases of the outputs
    basis - determines how many rows can possibly be in the output
            e.g. ['atom'] means one row per atom
                 ['element','struct'] means one row per (element,struct) pair
    constr - expression which must be evaluated as true in WHERE clause
    aconstr- expression which must be evaluated as true in HAVING clause
             (to do, automatically distinguish constr
             from aconstr by whether or not contains any aggs?)
    option - Objects which may or may not exist (i.e. LEFT JOIN on these)
    opt_attr - Attributes mentioned in query which may be null
            (otherwise NOT NULL constraint added to WHERE)
    """

    def __init__(
        self,
        exprs: M[str, Expr],
        basis: L[U[str, Entity]] = None,
        constr: Expr = None,
        aggcols: Sequence[Expr] = None,
        aconstr: Expr = None,
        option: Sequence[RelTup] = None,
        opt_attr: Sequence[PathAttr] = None,
    ) -> None:
        err = "Expected %s, but got %s (%s)"
        for k, v in exprs.items():
            assert isinstance(k, str), err % ("str", k, type(k))
            assert isinstance(v, Expr), err % ("Expr", v, type(v))

        self.exprs = exprs
        self.aggcols = aggcols or []
        self.constr = constr or true
        self.aconstr = aconstr or None

        if not basis:
            attrobjs = nub([a.obj for a in self.allattr()], str)
            assert len(attrobjs) == 1, f"Cannot guess basis for you {attrobjs}"
            self.basis = attrobjs
        else:
            self.basis = [x if isinstance(x, str) else x.name for x in basis]

        self.option = option or []

        if opt_attr:
            for a in opt_attr:
                if isinstance(a, PK):
                    raise ValueError("You can' have an optional ID attrs currently")
                assert isinstance(a, PathAttr), err % ("PathAttr", a, type(a))
        self.opt_attr = opt_attr or []

    def __str__(self) -> str:
        return "Query<%d exprs>" % (len(self.exprs))

    def __getitem__(self, key: str) -> Arg:
        err = "%s not found in query exprs %s"
        assert key in self.exprs, err % (key, self.exprs)
        return Arg(self.hash, key)

    ####################
    # Abstract methods #
    ####################
    def show(self, f: Fn) -> str:
        """How to represent a Query as a subselect"""
        raise NotImplementedError

    def fields(self) -> L[Expr]:
        """Might be missing a few things here .... """
        return list(self.exprs.values())

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(
            cls,
            exprs=dictionaries(keys=nonempty, values=Expr._strat()),
            basis=lists(nonempty, min_size=1, max_size=2),
            constr=Expr._strat(),
            aggcols=lists(Expr._strat(), max_size=2),
            aconstr=Expr._strat(),
            option=lists(RelTup._strat(), max_size=2),
            opt_attr=lists(PathAttr._strat(), max_size=2),
        )

    ####################
    # Public methods #
    ####################

    def allobj(self) -> L[str]:
        """All object names that are mentioned in query"""

        for a in self.allattr():
            if not hasattr(a, "obj"):
                raise ValueError(f"{a} is missing attr obj")
        return [o.obj for o in self.option] + self.basis + [a.obj for a in self.allattr()]

    def allattr(self) -> S[PathAttr]:
        """All path+attributes that are mentioned in query"""
        agg_x = [expr_attrs(ac) for ac in self.aggcols]  # type: L[L[PathAttr]]
        es = list(self.exprs.values()) + [self.constr, self.aconstr or true]
        out = set(flatten([expr_attrs(expr) for expr in es] + agg_x))
        return out

    def allrels(self) -> S[RelTup]:
        """All relations EXPLICITLY mentioned in the query"""
        out = set(self.option)
        for a in self.allattr():
            out = out | a.allrels()
        return out

    ###################
    # Private methods #
    ###################
    def _make_from(self) -> From:
        """ FROM clause of a query - combining FROMs from all the Paths """
        f = From(self.basis)
        attrs = self.allattr()
        for a in attrs:
            f = f | a.path._from()
        return f

    def showQ(
        self,
        not_deleted: bool = False,
        not_null: bool = True,
        limit: int = None,
    ) -> str:
        """
        Render a query

        To do: HAVING Clause
        """
        # FROM clause
        # ------------
        f = self._make_from()

        # What we select for
        # -------------------
        cols = ",\n\t".join([f'{e} AS "{k}"' for k, e in self.exprs.items()])  # .show(shower)
        cols = "" + cols if cols else ""
        # WHERE and HAVING clauses
        # ---------------------------------
        # Aggregations refered to in WHERE are treated specially
        where = str(self.constr)  # self.show_constr({})
        notdel = "\n\t".join([f'AND NOT COALESCE("{o}".deleted, False)' for o in f.aliases()])
        notnul = "\n\t".join([f"AND {x} IS NOT NULL" for x in self.allattr() if x not in self.opt_attr])
        where_args = [where]
        where_args += [notdel] if not_deleted else []
        where_args += [notnul] if not_null else []

        consts = "WHERE %s" % ("\n\t".join(where_args))

        # HAVING aggregations are treated 'normally'
        if self.aconstr:
            haves = f"\nHAVING {self.aconstr}"
        else:
            haves = ""

        # Group by clause, if anything SELECT'd has an aggregation
        # ---------------------------------------------------------
        def gbhack(x: Expr) -> str:
            """We want to take the CONCAT(PK," ",UID) exprs generated by Entity.id()
            and just replace them with PK"""
            if isinstance(x, PK):
                return str(x.pk)
            else:
                return str(x)

        gb = [gbhack(x) for x in self.aggcols]
        groupby = "\nGROUP BY \n\t%s" % (",\n\t".join(gb)) if gb else ""

        # Compute FROM clause
        # --------------------
        f_str = f.print(self.option)

        # Get the Limit Clause if limit is set
        # ------------------------------------
        limit_str = f"\nLimit {limit}" if limit is not None else ""
        # Put everything together to make query string
        # ----------------------------------------------------
        fmt_args = [cols, f_str, consts, groupby, haves, limit_str]
        output = "SELECT \n\t{}\n{}\n{}{}{}{}".format(*fmt_args)

        return output

    def get_row_count(self, db: ConnI) -> int:
        """
        Queries the database using the connection to get the number of rows this
        query is expected to return.

        Args:
            db (ConnI): ConnectInfo object of the database.

        Raises:
            ValueError: If the query doesn't return an output

        Returns:
            int: number of rows the query will return.
        """
        prepend = """SELECT\n\tCOUNT(*)\nFROM ("""
        append = """\n) AS X"""
        statement = prepend + self.showQ() + append
        query_output = select_dict(db, statement)
        if query_output:
            row_count = query_output[0][0]
            return row_count
        raise ValueError(f"No query output for row_count query,\n{statement}")

    def exec_query(self, db: ConnI) -> L[dict]:
        """Execute a query object in a giving model using a database connection"""
        return select_dict(db, self.showQ())



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Exceptions used by DBgen"""
from hypothesis.strategies import SearchStrategy, builds
from psycopg2.errors import Error as Psycopg2Error  # type: ignore  # noqa: F401

# Internal Imports
from dbgen.utils.misc import Base


class DBgenException(Exception, Base):
    """
    Base class for all DBgen's errors.
    Each custom exception should be derived from this class
    """

    status_code = 500

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenExternalError(DBgenException):
    """Custom class for catching errors that occur in code external to dbgen"""

    status_code = 400

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenInternalError(DBgenException):
    """
    Custom class for catching errors that occur in code internal to dbgen
    """

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenGeneratorError(DBgenException):
    """Custom class for catching errors that occur in code external to dbgen"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenSkipException(DBgenException):
    """Custom class for errors raised when a generator is skipped"""

    def __init__(self, msg) -> None:
        super(self.__class__, self).__init__()
        self.msg = msg

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenNotFoundException(DBgenException):
    """Raise when the requested object/resource is not available in the system"""

    status_code = 404

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenConfigException(DBgenException):
    """Raise when there is configuration problem"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenLoadException(DBgenException):
    """Raise when there is loading problem"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenMissingInfo(DBgenException):
    """Raise when there is missing info in a class/function initialization"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenInvalidArgument(DBgenException):
    """Raise when there is missing info in a class/function initialization"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


class DBgenTypeError(DBgenException):
    """Raise when a value cannot be cast the correct type"""

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Internal
from dbgen import Model
from dbgen.example.generators.analysis import analysis
from dbgen.example.generators.io import io
from dbgen.example.schema import all

################################################################################


def make_model() -> Model:
    # Initialize model
    m = Model("example")

    # Add objects and relations
    m.add(all)  # type: ignore

    # Add Generators
    io(m)
    analysis(m)
    return m



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from hashlib import sha256
from json import dumps
from typing import Any

Any
################################################################################
# Number of hash values
HASH_COUNT = 18446744073709551616


def hash_(x: "Any") -> str:
    """
    NEW: Take a list of showable things and generate a unique hash value in
       longint range (-9223372036854775808 to +9223372036854775807)
    """
    json_string = dumps(x, sort_keys=True, indent=4, separators=(",", ": "))
    return hashdata_(json_string)


def hashdata_(input_data: "Any") -> str:
    """
    OLD VERSION: Creates a 128 Byte hash value #return sha512(str(x).encode()).hexdigest()
    NEW: Take a list of showable things and generate a unique hash value in
       longint range (-9223372036854775808 to +9223372036854775807)
    """
    encoded_string = str(input_data).encode("utf-8")
    hex_hash = sha256(encoded_string).hexdigest()
    int_hash = int(hex_hash, 16)
    converted_val = str((int_hash % HASH_COUNT) - int(HASH_COUNT / 2))
    return converted_val


def abbreviate(x: "Any") -> "Any":
    """Truncate super long messages."""
    if isinstance(x, str) and len(x) > 1000:
        return x[:1000] + "...\n\t"
    else:
        return x


def cap(x: str) -> str:
    """Capitalize a string."""
    return x[0].upper() + x[1:]


def levenshteinDistance(s1: str, s2: str) -> int:
    """Distance between strings: # of characters difference"""
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = list(range(len(s1) + 1))
    for i2, c2 in enumerate(s2):
        distances_ = [i2 + 1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import defaultdict
from typing import TYPE_CHECKING, Any
from typing import Dict as D
from typing import List as L
from typing import Optional as Opt
from typing import Set as S
from typing import Tuple as T

import hypothesis.strategies as st
from hypothesis.strategies import SearchStrategy, builds, lists, text

from dbgen.core.func import Env, Func
from dbgen.core.funclike import Arg, PyBlock
from dbgen.core.load import Load
from dbgen.core.misc import Dep
from dbgen.core.query import Query
from dbgen.core.schema import Entity
from dbgen.utils.exceptions import DBgenSkipException
from dbgen.utils.lists import concat_map
from dbgen.utils.misc import Base
from dbgen.utils.sql import Connection as Conn
from dbgen.utils.sql import DictCursor, mkInsCmd, mkSelectCmd, mkUpdateCmd, sqlexecute, sqlselect
from dbgen.utils.str_utils import hash_

# Internal
if TYPE_CHECKING:
    from dbgen.core.model.model import Model

    Model
"""
Defines a Generator, as well as a Model method that is directly related
"""
################################################################################


class Generator(Base):
    """
    Generator: populates database with data
    One of the two component objects of a DBgen model
    """

    def __init__(
        self,
        name: str,
        desc: str = None,
        query: Query = None,
        transforms: L[PyBlock] = None,
        loads: L[Load] = None,
        tags: L[str] = None,
        batch_size: int = None,
        additional_deps: Dep = None,
    ) -> None:
        """
        Initializes generator object.

        [Query][dbgen.core.query.Query]

        Args:
            name (str): name of the generator
            desc (str, optional): Description of what the generator does. Defaults to None.
            query (Query, optional): [Query][dbgen.core.query.Query] object. Defaults to None.
            transforms (L[PyBlock], optional): [description]. Defaults to None.
            loads (L[Load], optional): [description]. Defaults to None.
            tags (L[str], optional): [description]. Defaults to None.
            env (Env, optional): [description]. Defaults to None.
            batch_size (int, optional): [description]. Defaults to None.
        """
        # assert loads, 'Cannot have generator which does nothing'
        assert name
        self.name = name.lower()
        self.desc = desc or "<no description>"
        self.query = query
        self.transforms = self._order_transforms(transforms or [], query)
        self.loads = loads or []
        self.tags = [t.lower() for t in tags or []]
        self.batch_size: Opt[int] = batch_size
        self.additional_deps = additional_deps
        super().__init__()

    def __str__(self) -> str:
        return f"Gen<{self.name}>"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        """A hypothesis strategy for generating random examples."""
        limited_text = text(min_size=1, max_size=2)
        return builds(
            cls,
            name=limited_text,
            desc=limited_text,
            query=st.one_of([st.none(), Query._strat()]),
            transforms=lists(PyBlock._strat(), max_size=2),
            loads=lists(Load._strat(), min_size=0, max_size=2),
            tags=lists(limited_text, max_size=3),
        )

    ##################
    # Public Methods #
    ##################

    def update_status(self, conn: Conn, run_id: int, status: str, err: str = "") -> None:
        """Update this gens status in the meta-database, also set error if provided"""
        cols = ["status", "error"] if err else ["status"]
        binds = [status, err, run_id, self.name] if err else [status, run_id, self.name]
        q = mkUpdateCmd("gens", cols, ["run", "name"])
        sqlexecute(conn, q, binds)

    def get_id(self, c: Conn) -> L[tuple]:  # THIS IS OBSOLETE BC HASH IS ID?
        """ Assuming we've inserted already """
        check = self.hash
        get_a = mkSelectCmd("gen", ["gen_id"], ["gen_id"])
        return sqlselect(c, get_a, [check])

    def hasher(self, x: Any) -> str:
        """Unique hash function to this Generator"""
        return hash_(str(self.hash) + str(x))

    def dep(self, universe: D[str, Entity]) -> Dep:
        """
        Determine the tabs/cols that are both inputs and outputs to the Gen

        Args:
            universe (D[str, Entity]): Mapping of object name to DBgen

        Returns:
            Dep
        """
        # Analyze allattr and allobj to get query dependencies
        if self.query:
            tabdeps = self.query.allobj()
            coldeps = [f"{a.obj}.{a.name}" for a in self.query.allattr()]
            for r in self.query.allrels():
                coldeps.append(r.obj + "." + r.rel)
        else:
            tabdeps, coldeps = [], []

        # Analyze loads to see what new cols and tabs are yielded
        newtabs: L[str] = []
        newcols: L[str] = []

        for a in self.loads:
            tabdeps.extend(a.tabdeps(universe))
            newtabs.extend(a.newtabs(universe))
            newcols.extend(a.newcols(universe))

        # Allow for unethical hacks
        for t in self.tags:
            if t[:4] == "dep ":
                coldeps.append(t[4:])

        implicit_deps = Dep(tabdeps, coldeps, newtabs, newcols)
        # check if we have additional explicit dependencies
        if self.additional_deps:
            total_deps = Dep.merge([implicit_deps, self.additional_deps])
        else:
            total_deps = implicit_deps
        return total_deps

    def add(self, cxn: "Conn") -> int:
        """
        Add the Generator to the metaDB which stores info about a model (if
        it's not already in there) and return the ID
        """
        a_id = self.get_id(cxn)
        if a_id:
            return a_id[0][0]
        else:
            cmd = mkInsCmd("gen", ["gen_id", "name", "description", "gen_json"])
            sqlexecute(cxn, cmd, [self.hash, self.name, self.desc, self.toJSON()])
            aid = self.get_id(cxn)
            return aid[0][0]

    def rename_object(self, o: Entity, n: str) -> "Generator":
        """Change all references to an object to account for name change"""
        g = self.copy()
        if g.query:
            g.query.basis = [n if b == o.name else b for b in g.query.basis]
        for i, a in enumerate(g.loads):
            g.loads[i] = a.rename_object(o, n)
        return g

    def purge(self, conn: Conn, mconn: Conn, universe: D[str, Entity]) -> None:
        """
        If a generator is purged, then any
        tables it populates will be truncated. Any columns it populates will be set all
        to NULL
        """
        d = self.dep(universe)
        tabs, cols = d.tabs_yielded, d.cols_yielded
        for t in tabs:
            sqlexecute(conn, f"TRUNCATE {t} CASCADE")

        for t, c in map(lambda x: x.split("."), cols):
            sqlexecute(mconn, f"UPDATE {t} SET {c} = NULL")

        gids = sqlselect(mconn, "SELECT gen_id FROM gen WHERE name = %s", [self.name])
        for gid in gids:
            sqlexecute(mconn, "DELETE FROM repeats WHERE gen_id = %s", [gid])

    def test(
        self,
        universe,
        input_rows: L[D[str, Any]],
        rename_dict: bool = True,
        verbose: bool = False,
    ) -> T[L[dict], L[D[str, L[D[str, Any]]]]]:
        # Apply the
        output_dicts = []
        for row in input_rows:
            result_dict = {self.query.hash: row} if self.query else {}
            try:
                if verbose:
                    from tqdm import tqdm

                    with tqdm(total=len(self.transforms)) as tq:
                        for pb in self.transforms:
                            tq.set_description(pb.func.name)
                            result_dict[pb.hash] = pb(result_dict)
                            tq.update()
                else:
                    for pb in self.transforms:
                        result_dict[pb.hash] = pb(result_dict)
            except DBgenSkipException as exc:
                result_dict["DBGEN_ERROR"] = exc.msg
                output_dicts.append(result_dict)
                continue
            # Replace pyblock hashes with function names if flag is True
            lambda_count = 0
            func_name_dict = {}
            name_count: D[str, int] = defaultdict(int)
            for pb in self.transforms:
                name = pb.func.name
                if pb.func.is_lam:
                    func_name_dict[pb.hash] = f"lambda{lambda_count}->{pb.outnames}"
                    lambda_count += 1
                else:
                    # Need to account for multiple pyblocks using same function
                    if name_count[name] > 0:
                        func_name_dict[pb.hash] = "_".join([name, str(name_count[name])])
                    else:
                        func_name_dict[pb.hash] = name
                    name_count[name] += 1

            output_dicts.append(result_dict)

        load_dicts: D[str, list] = defaultdict(list)
        for i, a in enumerate(self.loads):
            output_dict = a.test(universe=universe, rows=output_dicts)
            for table_name, rows in output_dict.items():
                load_dicts[table_name].extend(rows)

        # Rename PyBlocks
        if rename_dict:
            for i, row in enumerate(output_dicts):
                output_dicts[i] = {func_name_dict.get(key, "query"): val for key, val in row.items()}

        return output_dicts, [load_dicts]

    def test_with_db(
        self,
        universe,
        db: Conn = None,
        limit: int = 5,
        rename_dict: bool = True,
        interact: bool = False,
        input_rows: L[dict] = [],
    ) -> T[L[D[str, dict]], L[D[str, L[dict]]]]:
        assert limit <= 200, "Don't allow for more than 200 rows with test with db"
        assert (
            db is not None or input_rows
        ) or self.query is None, "Need to provide a db connection if generator has a query"

        if db is not None and self.query is not None:
            cursor = db.connect(auto_commit=False).cursor(f"test-{self.name}", cursor_factory=DictCursor)
            # If there is a query get the row count and execute it
            query_str = self.query.showQ(limit=limit)
            print("Executing Query...")
            cursor.execute(query_str)
            input_rows.extend(cursor.fetchall())
            print("Fetching Rows...")
            print("Closing Connection...")
            cursor.close()

        if interact:
            from dbgen.utils.interact import interact_gen

            return interact_gen(universe, self, input_rows)
        else:
            if self.query is None and len(input_rows) == 0:
                input_rows = [{}]

            return self.test(universe, input_rows, rename_dict)

    @property
    def env(self) -> Env:
        self._env = Env()
        for pyblock in self.transforms:
            self._env += pyblock.func.env
        return self._env

    ##################
    # Private Methods #
    ##################

    def _constargs(self) -> L[Func]:
        return concat_map(lambda y: y._constargs(), self.transforms)

    @staticmethod
    def _order_transforms(pbs: L[PyBlock], q: Opt[Query]) -> L[PyBlock]:
        """Make dependency graph among PyBlocks and determine execution order"""
        from dbgen.utils.graphs import DiGraph, topsort_with_dict

        G = DiGraph()
        d = {pb.hash: pb for pb in pbs}
        G.add_nodes_from(d.keys())
        for pb in pbs:
            for arg_ind, a in enumerate(pb.args):
                if isinstance(a, Arg) and (not q or a.key != q.hash):
                    if a.key not in d:
                        raise KeyError(
                            f"Argument {arg_ind} of {pb.func.name} refers to an object with a hash key {a.key} asking for name \"{getattr(a,'name','<No Name>')}\" that does not exist in the namespace."
                            "Did you make sure to include all PyBlocks in the func kwarg of Generator()?"
                        )
                    assert a.key in d, pb.func.name
                    G.add_edge(a.key, pb.hash)
        return topsort_with_dict(G, d)

    def _get_all_saved_key_dict(self) -> D[str, S[str]]:
        saved_keys = {}  # type: D[str,S[str]]
        for act in self.loads:
            for hash_loc, name_set in self._get_saved_key_dict(act).items():
                saved_keys.update({hash_loc: {*name_set, *saved_keys.get(hash_loc, set())}})

        return saved_keys

    def _get_saved_key_dict(self, load: Load) -> D[str, S[str]]:
        saved_keys = {}  # type: D[str,S[str]]
        if load.pk:
            if isinstance(load.pk, Arg):
                hash_loc = load.pk.key
                arg_name = load.pk.name
                saved_keys.update({hash_loc: {arg_name, *saved_keys.get(hash_loc, set())}})

        for val in load.attrs.values():
            if isinstance(val, Arg):
                saved_keys.update({val.key: {val.name, *saved_keys.get(val.key, set())}})

        for fk_load in load.fks.values():
            for hash_loc, name_set in self._get_saved_key_dict(fk_load).items():
                saved_keys.update({hash_loc: {*name_set, *saved_keys.get(hash_loc, set())}})

        return saved_keys

    # ######################
    # # Airflow Operator Exports
    # # --------------------
    def operator(self, model_name: str, run: int, universe: D[str, Entity]) -> str:

        # Get the necessary template
        from dbgen.templates import jinja_env

        gen_template = jinja_env.get_template("generator.py.jinja")

        # Prepare the rendered arguments
        pbs = [("pb" + str(pb.hash).replace("-", "neg"), pb.hash, pb.make_src()) for pb in self.transforms]
        loaders = [loader.make_src() for loader in self.loads]
        objs = {oname: (o.id_str, repr(o.ids()), repr(o.id_fks())) for oname, o in universe.items()}
        consttransforms = [cf.src for cf in self._constargs()]

        # Set the template arguements
        template_kwargs = dict(
            name=self.name,
            pyblocks=pbs,
            genname=self.name,
            loads=loaders,
            objs=objs,
            query=self.query.showQ() if self.query else False,
            queryhash=self.query.hash if self.query else None,
            run=run,
            model_name=model_name,
            consttransforms=consttransforms,
        )

        return gen_template.render(**template_kwargs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
from os import environ
from os.path import join

# External imports
from typing import TYPE_CHECKING, Any
from typing import List as L

# Internal Imports
if TYPE_CHECKING:
    from dbgen.core.model.model import Model

# from dbgen.core.misc import ConnectInfo
from dbgen.templates import jinja_env
from dbgen.utils.config import DBGEN_HOME, config


##################################
def run_airflow(
    self: "Model",
    sched: str = "@once",
    nuke: bool = False,
    start: str = None,
    until: str = None,
    xclude: L[str] = [],
    only: L[str] = [],
    retry: bool = False,
    serial: bool = False,
    bar: bool = True,
    clean: bool = False,
    batch: int = None,
    **kwargs: Any,
) -> None:
    """
    Create an airflow DAG, then execute it.
    """
    # Validate input
    startErr = 'Starting generator ("start") must be a Generator name'
    assert not start or start in self.gens, startErr
    tillErr = 'Final generator ("until") must be a Generator name'
    assert not until or until in self.gens, tillErr
    xclude_ = set(xclude)
    only_ = set(only)
    for w in only_ | xclude_:
        self._validate_name(w)

    # from airflow.hooks.postgres_hook import PostgresHook

    # ping the database and check if we need to add Objs and Relations
    # connection = PostgresHook.get_connection(self.name)
    # connI = ConnectInfo.from_postgres_hook(connection)
    # mconnection = PostgresHook.get_connection(self.name + "_log")
    # mconnI = ConnectInfo.from_postgres_hook(mconnection)
    # if nuke:
    #     self.make_schema(conn=connI, nuke=nuke)  # FULL NUKE
    # # Check if the schema exists
    # if not self.check_schema_exists(connI):
    #     raise ValueError("Your Schema doesn't exist yet, please run with --nuke=T the first time")
    # Make metatables
    # ----------------
    # run_id = self._make_metatables(
    #     mconn=mconnI,
    #     conn=connI,
    #     nuke=nuke,
    #     retry=False,
    #     only=sorted(only_),
    #     xclude=sorted(xclude_),
    #     start=start,
    #     until=until,
    #     bar=False,
    # )
    run_id = 0
    gen_hash_dict = {gen_name: gen.hash for gen_name, gen in self.gens.items()}
    objs = {oname: (o.id_str, repr(o.ids()), repr(o.id_fks())) for oname, o in self.objs.items()}
    deps = list(self._gen_graph().edges())
    template_kwargs = dict(
        user=environ.get("USER", "dbgen"),
        objs=objs,
        modelname=self.name,
        run_id=run_id,
        retry=retry,
        serial=serial,
        bar=bar,
        user_batch_size=batch,
        gen_hash_dict=gen_hash_dict,
        deps=deps,
        schedule_interval=sched,
        date=datetime.date(datetime.now()),
        operator_name="DummyGenOperator",
    )

    dag_template = jinja_env.get_template("run_airflow.py.jinja")
    dag_file_contents = dag_template.render(**template_kwargs)

    # Write the contents of the dag file to the
    DAG_FOLDER = config.get("core", "DAG_FOLDER", fallback=join(DBGEN_HOME, "dags/"))
    new_dag_file_pth = join(DAG_FOLDER, "test.py")
    with open(new_dag_file_pth, "w") as f:
        f.write(dag_file_contents)
    # if not exists(join(DAG_FOLDER,str(self.hash)+'.py')) or True:
    #     # move old model to archive
    #     dag_files = glob(join(DAG_FOLDER,'*.py'))
    #     if dag_files:
    #         assert len(dag_files) == 1, f'Why are is more than one file in the Dag folder: {DAG_FOLDER}?'
    #         old_dag_file_pth = dag_files[0]
    #         shutil.move(old_dag_file_pth,join(DAG_FOLDER,'archive'))
    #     new_dag_file_pth = join(DAG_FOLDER, str(self.hash)+'.py')
    #     with open(new_dag_file_pth, 'w') as f:
    #         f.write(dag_file_contents)

    print("Run Airflow!")



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List as L
from typing import Type

from hypothesis.strategies import SearchStrategy, one_of, recursive

from dbgen.core.expr.expr import (
    ABS,
    AND,
    AVG,
    BINARY,
    COALESCE,
    CONCAT,
    COUNT,
    DIV,
    EQ,
    GE,
    GT,
    JSON_EXTRACT,
    LE,
    LEFT,
    LEN,
    LIKE,
    LT,
    MAX,
    MIN,
    MINUS,
    MUL,
    NE,
    NOT,
    NULL,
    OR,
    PLUS,
    POW,
    REGEXP,
    RIGHT,
    SQRT,
    SUM,
    Binary,
    Literal,
    Nary,
    Tup,
    Unary,
)
from dbgen.core.expr.pathattr import PathAttr


# jsonstrat = recursive(none() | booleans() | floats() | text(),
#                       lambda children: lists(children, 1) |
#                       dictionaries(text(), children, min_size=1))
def nullarystrat() -> SearchStrategy:
    return one_of(Literal._strat(), PathAttr._strat())


def unarystrat(x: SearchStrategy) -> SearchStrategy:
    uns = [
        ABS,
        SQRT,
        MAX,
        SUM,
        MIN,
        AVG,
        COUNT,
        BINARY,
        LEN,
        NOT,
        NULL,
    ]  # type: L[Type[Unary]]
    return one_of(*[u._strat(x) for u in uns])


def binarystrat(x: SearchStrategy) -> SearchStrategy:
    bins = [
        REGEXP,
        LIKE,
        MUL,
        DIV,
        PLUS,
        MINUS,
        POW,
        LEFT,
        RIGHT,
        JSON_EXTRACT,
        EQ,
        NE,
        LT,
        GT,
        LE,
        GE,
    ]  # type: L[Type[Binary]]
    return one_of(*[b._strat(x) for b in bins])


def narystrat(x: SearchStrategy) -> SearchStrategy:
    ns = [CONCAT, COALESCE, Tup, OR, AND]  # type: L[Type[Nary]]
    return one_of(*[n._strat(x) for n in ns])


def exprstrat() -> SearchStrategy:
    return recursive(nullarystrat(), lambda x: unarystrat(x) | binarystrat(x) | narystrat(x))


if __name__ == "__main__":
    ns = nullarystrat().example()
    us = unarystrat(nullarystrat()).example()
    bs = binarystrat(nullarystrat()).example()
    Ns = narystrat(nullarystrat()).example()
    uus = unarystrat(unarystrat(nullarystrat())).example()



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from argparse import ArgumentParser
from distutils.util import strtobool
from re import DOTALL, MULTILINE, finditer, search
from typing import Optional, Tuple


################################################################################
def parse_line(string: str, substr: str, index: int = 0) -> Optional[str]:
    """
    Returns the n'th line containing substring
    Any negative index will return last one.
    """
    iter = finditer(substr + ".*$", string, MULTILINE)
    found = False

    for match in iter:
        if index == 0:
            return match[0]
        else:
            index -= 1
            found = True
    if found:
        return match[0]  # negative input for index
    else:
        return None


def btw(s: str, begin: str, end: str, off: int = 0) -> Tuple[str, int]:
    result = search(f"{begin}(.*?){end}", s[off:], DOTALL)
    if result:
        if result.group(1) is None:
            raise ValueError(f"No Match: {s}")
        return result.group(1), result.end() + off
    else:
        return "", 0


def input_to_level(logging_level: str) -> int:
    """
    Convert logging level string to logging.LEVEL int.

    Args:
        logging_level (str): Logging level string (INFO/DEBUG/WARNING/CRITICAL)

    Raises:
        ValueError: when input string is not a valid logging level

    Returns:
        int: Gets the valid integer from the logging module
    """
    log_map = {
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "WARNING": logging.WARNING,
        "CRITICAL": logging.CRITICAL,
    }
    if logging_level in log_map:
        return log_map[logging_level]
    raise ValueError("Please provide a valid logging level")


########
# Command line parsing
parser = ArgumentParser(description="Run a DBG update", allow_abbrev=True)

parser.add_argument(
    "--nuke",
    default="",
    type=str,
    help="Reset the DB - needed if you make schema changes",
)

parser.add_argument(
    "--add",
    action="store_true",
    help="Try to add columns (if you make benign schema additions)",
)

parser.add_argument("--only", default="", help="Run only the (space separated) generators/tags")

parser.add_argument(
    "--xclude",
    type=str,
    default="",
    help="Run only the (space separated) generators/tags",
)

parser.add_argument("--start", default="", help="Start at the designed Generator")

parser.add_argument("--until", default="", help="Stop at the designed Generator")

parser.add_argument(
    "--retry",
    action="store_true",
    help="Ignore repeat checking",
)

parser.add_argument(
    "--serial",
    default=False,
    type=lambda x: bool(strtobool(x)),
    help='Ignore any "parallel" flags',
)

parser.add_argument(
    "--clean",
    default=False,
    help="Clean the database of the deleted column!DOESN't DELETE DELETED ROWS YET!",
)

parser.add_argument(
    "--skip-row-count",
    action="store_true",
    help="Skip Row count for large queries",
)

parser.add_argument(
    "--batch",
    default=None,
    type=lambda x: int(float(x)),
    help="Set default batch_size",
)

parser.add_argument(
    "--write-logs",
    action="store_true",
    help="write logs to local file",
)

parser.add_argument(
    "--log-level",
    default=logging.DEBUG,
    type=input_to_level,
    help="Set default batch_size",
)

parser.add_argument(
    "--log-path",
    default=None,
    help="Set path of output log file",
)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod
from ast import literal_eval
from inspect import _empty  # type: ignore

# External Modules
from typing import List as L
from typing import Type
from typing import TypeVar as TV

from hypothesis.strategies import SearchStrategy, builds, just, lists, one_of

# Internal modules
from dbgen.utils.misc import Base

"""
Defines the abstract class DataType and all classes that implement it
"""
################################################################################
class DataType(Base, metaclass=ABCMeta):
    """
    Internal representation of Python datatypes

    This is needed because the output datatypes from python's 'inspect' module
    are not serializable. To convert such an output into an instance of this
    class, use the class method DataType.get_datatype()

    This class is meant to be extended as new cases appear
    """

    def __init__(self) -> None:
        super().__init__()

    def __len__(self) -> int:
        """Default length, ONLY Tuple has a non-one length."""
        return 1

    @classmethod
    def _strat(cls) -> SearchStrategy:
        dts = [
            AnyType(),
            NoneType(),
            BaseType("x"),
            TypeVar("x"),
            Callable([AnyType()], NoneType()),
            Union([AnyType(), BaseType("x")]),
            Tuple([AnyType(), BaseType("x")]),
            List(NoneType()),
            Dict(BaseType("x"), BaseType("y")),
        ]  # type: L[DataType]
        return one_of(*[just(x) for x in dts])

    @abstractmethod
    def __str__(self) -> str:
        """Need to provide a str representation"""
        raise NotImplementedError

    @classmethod
    def get_datatype(cls, t: Type) -> "DataType":
        """
        Convert a Python Type object into a simple representation

        This is very ad hoc and hacky
        """
        if isinstance(t, str):
            try:
                t = literal_eval(t)  # try to evaluate, likely will fail
            except ValueError:
                pass

        strt = str(t)

        # Python builtins
        if t == float:
            return BaseType("Float")
        elif t == int:
            return BaseType("Int")
        elif t == str:
            return BaseType("Str")
        elif t == bool:
            return BaseType("Bool")
        elif isinstance(t, type(None)):
            return NoneType()

        # Special cases from chemistry libraries
        elif "Atoms" in strt:
            return BaseType("Atoms")
        elif strt == "Structure":
            return BaseType("Structure")
        elif strt in ["BULK", "<class 'bulk_enumerator.bulk.BULK'>"]:
            return BaseType("BULK")
        elif "Graph" in strt:
            return BaseType("MultiGraph")
        elif "Structure" in strt:
            return BaseType("Structure")

        # Typing higher order types
        elif str(t.__class__) == str(Union) or strt[:12] == "typing.Union":
            return Union([cls.get_datatype(x) for x in t.__args__])

        elif "Tuple" in strt:
            return Tuple([cls.get_datatype(x) for x in t.__args__])

        elif "Callable" in strt:
            from_args = [cls.get_datatype(x) for x in t.__args__[:-1]]
            to_arg = cls.get_datatype(t.__args__[-1])
            return Callable(from_args, to_arg)

        elif isinstance(t, TV):  # type: ignore
            return TypeVar(t.__name__)

        elif strt[:11] == "typing.List":
            return List(cls.get_datatype(t.__args__[0]))

        elif strt[:11] == "typing.Dict":
            a1, a2 = [cls.get_datatype(x) for x in t.__args__]
            return Dict(a1, a2)

        elif strt == "typing.Any" or t == _empty:  # type: ignore
            return AnyType()

        else:
            raise NotImplementedError(f"NEW DATATYPE FOUND {strt}")


################################################################################


class AnyType(DataType):
    def __str__(self) -> str:
        return "Any"
        super().__init__()

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


################################################################################


class NoneType(DataType):
    def __str__(self) -> str:
        return "None"
        super().__init__()

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


################################################################################


class BaseType(DataType):
    def __init__(self, unBase: str) -> None:
        self.unBase = unBase
        super().__init__()

    def __str__(self) -> str:
        return f'"{self.unBase}"'

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


################################################################################


class TypeVar(DataType):
    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__()

    def __str__(self) -> str:
        return f'"{self.name}"'

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


################################################################################


class Callable(DataType):
    def __init__(self, c_args: L[DataType], out: DataType) -> None:
        self.c_args = c_args
        self.out = out
        super().__init__()

    def __str__(self) -> str:
        ar = " -> "
        return ar.join(map(str, self.c_args)) + ar + str(self.out)

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(
            cls,
            out=DataType._strat(),
            c_args=lists(DataType._strat(), min_size=1, max_size=2),
        )


################################################################################


class Union(DataType):
    def __init__(self, args: L[DataType]) -> None:
        self.args = args
        super().__init__()

    def __str__(self) -> str:
        return "{%s}" % (",".join(sorted(map(str, self.args))))

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls, args=lists(DataType._strat(), min_size=1, max_size=2))


################################################################################


class Tuple(DataType):
    def __init__(self, args: L[DataType]) -> None:
        self.args = args
        super().__init__()

    def __str__(self) -> str:
        return f"({','.join(map(str, self.args))})"

    def __len__(self) -> int:
        return len(self.args)

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls, args=lists(DataType._strat(), min_size=1, max_size=2))


################################################################################


class List(DataType):
    def __init__(self, content: DataType) -> None:
        self.content = content
        super().__init__()

    def __str__(self) -> str:
        return f"[{str(self.content)}]"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)


################################################################################


class Dict(DataType):
    def __init__(self, key: DataType, val: DataType) -> None:
        self.key = key
        self.val = val
        super().__init__()

    def __str__(self) -> str:
        return f"{{ {str(self.key)} : {str(self.val)} }}"

    @classmethod
    def _strat(cls) -> SearchStrategy:
        return builds(cls)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from importlib import import_module


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError:
        raise ImportError(f"{dotted_path} doesn't look like a module path")

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class')



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod
from datetime import datetime
from random import choice
from re import split
from string import ascii_lowercase, ascii_uppercase, digits

# External Modules
from typing import Any, Optional

from dbgen.utils.exceptions import DBgenTypeError
from dbgen.utils.misc import Base

"""
Representations of SQL Data Types
"""
################################################################################
chars = ascii_lowercase + ascii_uppercase + digits


class SQLType(Base, metaclass=ABCMeta):
    """
    SQL datatypes
    """

    data = {}  # type: dict

    @abstractmethod
    def __str__(self) -> str:
        """String representation to be used in raw SQL expression"""
        raise NotImplementedError

    @abstractmethod
    def cast(self, val):
        """Cast a value as SQLType"""
        raise NotImplementedError

    def __init__(self) -> None:
        pass

    @staticmethod
    def from_str(s: str) -> "SQLType":
        """
        Ad hoc string parsing
        """
        if "VARCHAR" in s:
            mem = split(r"\(|\)", s)[1]
            return Varchar(int(mem))
        elif "DECIMAL" in s:
            prec, scale = split(r"\(|\)|,", s)[1:3]
            return Decimal(int(prec), int(scale))
        elif "INT" in s:
            if "TINY" in s:
                kind = "small"
            elif "BIG" in s:
                kind = "big"
            else:
                kind = "medium"
            signed = "UNSIGNED" not in s
            return Int(kind, signed)
        elif "TEXT" in s:
            if "TINY" in s:
                kind = "tiny"
            elif "MED" in s:
                kind = "medium"
            elif "LONG" in s:
                kind = "long"
            else:
                kind = ""
            return Text(kind)
        else:
            raise NotImplementedError("New SQLtype to parse? " + s)


class Varchar(SQLType):
    def __init__(self, mem: int = 255) -> None:
        self.mem = mem

    def __str__(self) -> str:
        return "VARCHAR(%d)" % self.mem

    def cast(self, val) -> Optional[str]:
        if val is None:
            return val
        return str(val)


class Decimal(SQLType):
    def __init__(self, prec: int = 15, scale: int = 6) -> None:
        self.prec = prec
        self.scale = scale

    def __str__(self) -> str:
        return "DECIMAL(%d,%d)" % (self.prec, self.scale)

    def cast(self, val) -> Optional[float]:
        if val is None:
            return val
        return float(val)


class Boolean(SQLType):
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return "Boolean"

    def rand(self) -> Any:
        return choice(["true", "false"])

    def cast(self, val) -> Optional[bool]:
        if val is None:
            return val
        elif not isinstance(val, bool):
            raise DBgenTypeError(f"Val is not a boolean! {val}")
        return bool(val)


class Int(SQLType):
    def __init__(
        self,
        kind: str = "medium",
        signed: bool = True,
    ) -> None:
        kinds = ["small", "medium", "big"]
        assert kind in kinds, f"Invalid Int type: {kind} not found in {kinds}"
        self.kind = kind
        self.signed = signed

    def __str__(self) -> str:
        options = ["small", "medium", "big"]
        if self.kind == "small":
            core = "SMALLINT"
        elif self.kind == "medium":
            core = "INTEGER"
        elif self.kind == "big":
            core = "BIGINT"
        else:
            err = 'unknown Int kind "%s" not in options %s '
            raise ValueError(err % (self.kind, options))
        return core + ("" if self.signed else " UNSIGNED")

    def cast(self, val) -> Optional[int]:
        if val is None:
            return val
        elif not isinstance(val, (int, float, str)) or isinstance(val, bool):
            raise DBgenTypeError(
                f"Val cannot be safely cast to int type: {val} {type(val)}\nSafe types are int, str, float."
            )
        return int(val)


class Text(SQLType):
    def __init__(self, kind: str = "") -> None:
        self.kind = kind

    def __str__(self) -> str:
        return "TEXT"

    def cast(self, val) -> Optional[str]:
        return str(val)


class Date(SQLType):
    def __str__(self) -> str:
        return "DATE"

    def cast(self, val) -> Optional[datetime]:
        if val is None:
            return val
        elif not isinstance(val, datetime):
            raise DBgenTypeError(f"Value cannot be safely cast to {str(self)}: {val}")
        return val


class Timestamp(SQLType):
    def __str__(self) -> str:
        return "TIMESTAMP"

    def cast(self, val) -> Optional[datetime]:
        if val is None:
            return val
        elif not isinstance(val, datetime):
            raise DBgenTypeError(f"Value cannot be safely cast to {str(self)}: {val}")
        return val


class Double(SQLType):
    def __str__(self) -> str:
        return "DOUBLE"

    def cast(self, val) -> Optional[float]:
        if val is None:
            return val
        elif not isinstance(val, float):
            raise DBgenTypeError(f"Value cannot be safely cast to {str(self)}: {val}")
        return val


class JSON(SQLType):
    def __str__(self) -> str:
        return "JSON"

    def cast(self, val) -> Optional[str]:
        if val is None:
            return val
        elif not isinstance(val, str):
            raise DBgenTypeError(f"Value cannot be safely cast to {str(self)}: {val}")
        return val


class JSONB(SQLType):
    def __str__(self) -> str:
        return "JSONB"

    def cast(self, val) -> Optional[str]:
        if val is None:
            return val
        elif not isinstance(val, str):
            raise DBgenTypeError(f"Value cannot be safely cast to {str(self)}: {val}")
        return val
